/*
    This file contains some abstract implementations of the Codec interface that make it simple to plug-in existing codecs

    Written by m^2.
    You can consider the mods to be public domain if you care to separate them.
    If your country doesn't recognize author's right to relieve themselves of copyright,
    you can use it under the terms of WTFPL version 2.0 or later.

*/

#include "abstractCodecs.hpp"
#include "tools.hpp"

#if __cplusplus >= 201103L // C++ 2011
    #include <cstdint>
#else
extern "C"
{
    #include <stdint.h>
}
#endif // C++ 2011
#include <sstream>
#include <iomanip>

using namespace std;

///////////////////////////
// Codec class
///////////////////////////


Codec::Codec(const std::string& name,
        const std::string& version,
        encoder_t encoder,
        decoder_t decoder,
        max_encoded_size_t max_size,
        transform_type encode_transform_type,
        transform_type decode_transform_type,
        bool can_be_skipped)
        throw(InvalidParams) :
    name(name),
    version(version),
    encoder(encoder),
    decoder(decoder),
    max_compressed_size(max_size),
    encode_transform_type(encode_transform_type),
    decode_transform_type(decode_transform_type),
    can_be_skipped(can_be_skipped)
{}
void Codec::init(const string& args, unsigned threads_no, size_t isize, bool init_encoder, bool init_decoder)
{
    this->args = args;
}
void Codec::cleanup()
{}

string Codec::help() const
{
    return this->name + " " + this->version + "\nTakes no parameters";
}

string Codec::introduction() const
{
    ostringstream oss(ostringstream::out);
    oss << setw(40) << left << this->name << setw(12) << this->version << " " << this->args;
    return oss.str();
}

Codec::InvalidParams::InvalidParams(const string& message) :
    message(message)
{}

size_t Codec::_max_compressed_size(size_t input_size)
{
    // I don't know the overhead of any individual codec that doesn't inform about it
    // So as a workaround I use these fairly-safe defaults
    return input_size + input_size/8 + 32;
}

void** Codec::eparams()
{
    return 0;
}
void** Codec::dparams()
{
    return 0;
}
///////////////////////////
// MultifunctionCodec
///////////////////////////

const CodecArgs& MultifunctionCodec::get_type(const string& args)
{
    for(size_t i=0; i<len; ++i)
    {
        if(case_insensitive_compare(allowed_args[i].args, args) == 0)
            return allowed_args[i];
    }
    throw InvalidParams(help());
}
const string& MultifunctionCodec::default_args(const string& args)
{
    return args == "" ? default_arg : args;
}


MultifunctionCodec::MultifunctionCodec(const string& name, const string& version, const CodecArgs* allowed_args, size_t len, const string& default_arg, max_encoded_size_t max_size) :
    Codec(name, version, 0, 0, max_size), // encoder and decoder are filled by init()
    allowed_args(allowed_args),
    len(len),
    default_arg(default_arg)
{}
void MultifunctionCodec::init(const string& args, unsigned threads_no, size_t isize, bool init_encoder, bool init_decoder)
{
    this->args = default_args(args);
    const CodecArgs& codec_args =  get_type(this->args);

    this->encoder = codec_args.encoder;
    this->decoder = codec_args.decoder;
}

string MultifunctionCodec::help() const
{
    string ret = this->name + " " + this->version + "\nAvaivable parameters: \n";
    for(size_t i=0; i<len; ++i)
    {
        if(! allowed_args[i].args.empty())
        {
            ret += " ";
            ret += allowed_args[i].args;
            ret += "\n";
        }
    }
    return ret;
}


///////////////////////////
// CodecWithIntModes
///////////////////////////
 
const string& CodecWithIntModes::default_args(const string& args) const
{
    return args == "" ? default_mode : args;
}

CodecWithIntModes::CodecWithIntModes(const string& name, const string& version,
        encoder_t encoder, decoder_t decoder,
        intptr_t min_mode, intptr_t max_mode, const string& default_mode,
        max_encoded_size_t max_size,
        transform_type encode_transform_type,
        transform_type decode_transform_type,
        bool can_be_skipped) :
    Codec(name, version, encoder, decoder, max_size, encode_transform_type, decode_transform_type, can_be_skipped),
    min_mode(min_mode),
    max_mode(max_mode),
    default_mode(default_mode)
{}
void CodecWithIntModes::init(const string& args, unsigned threads_no, size_t isize, bool init_encoder, bool init_decoder)
{
    this->args = default_args(args);
    intptr_t mode;
    try
    {
        from_string(this->args, mode);
    }
    catch(const bad_cast &bc)
    {
        throw InvalidParams(help());
    }
    if(mode < min_mode || mode > max_mode)
        throw InvalidParams(help());
    this->params = new void*[threads_no];
    for(unsigned i=0; i<threads_no; ++i)
    {
        this->params[i] = (void*)mode;
    }
}

void CodecWithIntModes::cleanup()
{
    delete[] this->params;
}

string CodecWithIntModes::help() const
{
    string ret;
    
    ret = this->name + " " + this->version + "\n";
    ret += "Available parameters: \n";
    ret += " Numbers from " + to_string(min_mode) + " to " + to_string(max_mode) + "\n";

    return ret;
}

void** CodecWithIntModes::eparams()
{
    return this->params;
}
void** CodecWithIntModes::dparams()
{
    return eparams();
}

///////////////////////////
// CombinationCodec
///////////////////////////

CombinationCodec::CombinationCodec(Codec& encoder, Codec& decoder):
    Codec(encoder.name    + "/" + decoder.name,
          encoder.version + "/" + decoder.version,
          encoder.encoder,
          decoder.decoder,
          encoder.max_compressed_size),
    encoderObject(encoder),
    decoderObject(decoder)
    {}

void CombinationCodec::init(const string& args, unsigned threads_no, size_t isize, bool init_encoder, bool init_decoder)
{
    string encoder_args;
    string decoder_args;

    string::size_type pos = args.find('/');
    if(pos == string::npos)
    {
        // either no parameters or parameters for encoder only
        // either way...
        encoder_args = args;
    }
    else
    {
        encoder_args = args.substr(0, pos);
        decoder_args = args.substr(pos + 1);
    }
    if(init_encoder)
        encoderObject.init(encoder_args, threads_no, isize, true, false);
    // A tricky part. 
    // To support "Codec/Codec,Params/Params2" properly,
    // That is: a single codec with separate params for compression and decompression,
    // I have to read encoder args before they are overwritten by decoder initiation
    // then initiate decoder and in turn read its params
    this->args = init_encoder ? encoderObject.args : encoder_args;

    if(init_decoder)
        decoderObject.init(decoder_args, threads_no, isize, false, true);

    this->encoder = encoderObject.encoder;
    this->decoder = decoderObject.decoder;

    this->args += "/";
    this->args += init_decoder ? decoderObject.args : decoder_args;

    this->encode_transform_type = init_encoder   ? encoderObject.encode_transform_type : in_place;
    this->decode_transform_type = init_decoder ? decoderObject.decode_transform_type : in_place;
}

void CombinationCodec::cleanup()
{
    encoderObject.cleanup();
    decoderObject.cleanup();
}

void** CombinationCodec::eparams()
{
    return encoderObject.eparams();
}
void** CombinationCodec::dparams()
{
    return decoderObject.dparams();
}