/*
    This file contains some abstract implementations of the Codec interface that make it simple to plug-in existing codecs

    Written by m^2.
    You can consider the mods to be public domain if you care to separate them.
    If your country doesn't recognize author's right to relieve themselves of copyright,
    you can use it under the terms of WTFPL version 2.0 or later.

*/

#ifndef ABSTRACTCODECS_HPP_B9fgkfG8
#define ABSTRACTCODECS_HPP_B9fgkfG8


#include "codecs.hpp"

#include <string>


///////////////////////////
// MultifunctionCodec
//
// Codecs that have several compression / decompression functions
// Each function pair is referred to by an argument
///////////////////////////
struct CodecArgs
{
    std::string args;
    encoder_t encoder;
    decoder_t decoder;
    CodecArgs(const std::string& args, encoder_t encoder, decoder_t decoder) :
        args(args),
        encoder(encoder),
        decoder(decoder)
    {}
};


struct MultifunctionCodec : Codec
{
private:

    const CodecArgs* allowed_args;
    size_t len;
    const std::string default_arg;

    const CodecArgs& get_type(const std::string& args);
    const std::string& default_args(const std::string& args);

public:

    MultifunctionCodec(const std::string& name, const std::string& version,
        const CodecArgs* allowed_args, size_t len, const std::string& default_arg,
        max_encoded_size_t max_size=_max_compressed_size);
        
    virtual std::string help() const;
    virtual void init(const std::string& args, unsigned threads_no, size_t isize, bool init_encoder=true, bool init_decoder=true);
    
};

///////////////////////////
// CodecWithIntModes
//
// Such codecs have just 1 compression and 1 decompression function
// that take ints as parameters to specify compression mode
///////////////////////////
struct CodecWithIntModes : Codec
{
private:

    const int min_mode;
    const int max_mode;
    const std::string default_mode;
    
    const std::string& default_args(const std::string& args) const;

public:

    CodecWithIntModes(const std::string& name, const std::string& version, 
        encoder_t encoder, decoder_t decoder, 
        intptr_t min_mode, intptr_t max_mode, const std::string& default_mode, 
        max_encoded_size_t max_size=_max_compressed_size,
        transform_type encode_transform_type=moving,
        transform_type decode_transform_type=moving,
        bool can_be_skipped=true);
        
    virtual std::string help() const;
    virtual void init(const std::string& args, unsigned threads_no, size_t isize, bool init_encoder=true, bool init_decoder=true);
    virtual void cleanup();
    
    virtual void** eparams();// parmeters to be sent to compressors
    virtual void** dparams();// parmeters to be sent to decompressors 

protected:
    void** params;
    
};

///////////////////////////
// CombinationCodec
//
// Takes encoder from 1 codec and decoder from the other
///////////////////////////
struct CombinationCodec : Codec
{
private:

    Codec& encoderObject;
    Codec& decoderObject;

public:

    CombinationCodec(Codec& encoder, Codec& decoder);
    
    // Initialises both codecs.
    // args takes form: [encoder_args[/decoder_args]]
    virtual void init(const std::string& args, unsigned threads_no, size_t isize, bool init_encoder=true, bool init_decoder=true);
    virtual void cleanup();
    
    virtual void** eparams();// parmeters to be sent to compressors
    virtual void** dparams();// parmeters to be sent to decompressors 
};
/////////////////////////////////////////////////
// Checksum
//
// Checksum is an in-place codec 
// that appends its digest right after the end of the data
/////////////////////////////////////////////////
template<size_t digest_size>
class Checksum : public Codec
{
    typedef void (*checksum_t)(char* in, size_t isize, char* out);
private:
    checksum_t checksummer;
    void** params;

public:
    Checksum(const std::string& name, const std::string& version, checksum_t checksummer) :
        Codec(name, version, encode, decode, max_output_size, Codec::in_place, Codec::in_place, false),
        checksummer(checksummer),
        params(0)
    {}

    static size_t max_output_size(size_t size) 
    {
        return size + digest_size;
    }
    static size_t encode(char* in, size_t isize, char* out, size_t osize, void* args)
    {
        checksum_t checksummer = *(checksum_t*) args;
        checksummer(in, isize, in + isize);
        return isize + digest_size;
    }
    static size_t decode(char* in, size_t isize, char* out, size_t osize, void* args)
    {
        checksum_t checksummer = *(checksum_t*) args;
        // there checksum of the original data is supposed to be stored right after the data
        size_t data_size = isize - digest_size;
        char digest[digest_size];
        checksummer(in, data_size, digest);
        if(memcmp(in + data_size, digest, digest_size) != 0)
            return CODING_ERROR;
        return data_size;
    }
    virtual void init(const std::string& args, unsigned threads_no, size_t isize, bool init_encoder=true, bool init_decoder=true)
    {
        if(!params)//TODO: Do other codecs handle correctly allocation when called first with init_encoder and then init_decoder?
        {
            params = new void*[threads_no];
            for(unsigned i=0; i < threads_no; ++i)
                params[i] = (void*)checksummer;
        }
    }
    virtual void cleanup()
    {
        if(params)
        {
            delete[] params;
            params = 0;
        }
    }
    virtual void** eparams()
    {
        return params;
    }
    virtual void** dparams()
    {
        return eparams();
    }
    
};

#endif // ABSTRACTCODECS_HPP_B9fgkfG8