/*
(C) 2011, Dell Inc. Written by Przemyslaw Skibinski (inikep@gmail.com)

    LICENSE

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 3 of
    the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details at
    Visit <http://www.gnu.org/copyleft/gpl.html>.

    Modified by m^2.
    You can consider the mods to be public domain if you care to separate them.
    If your country doesn't recognize author's right to relieve themselves of copyright,
    you can use it under the terms of WTFPL version 2.0 or later.

*/

///////////////////////////
// DEFINITIONS
///////////////////////////

#define PROGNAME "fsbench"
#define PROGVERSION "0.12"

///////////////////////////
// INCLUDES
///////////////////////////


#include "benchmark.hpp"
#include "threads.hpp"
#include "tools.hpp"

#include <climits>
#include <cstdlib>
#include <cstring>
#include <fstream>
#include <iostream>
#include <sstream>

#if defined(WIN32) || defined(_WIN32) || defined(__WIN32__) || defined(_WIN64)
    #include <windows.h>
    #define InitTimer(x) if (!QueryPerformanceFrequency(&x)) { cout<<"QueryPerformance not present"; };
    #define GetTime(x) QueryPerformanceCounter(&x);
#else
    #include <time.h>
    typedef struct timespec LARGE_INTEGER;
    #define InitTimer(x)
    #define GetTime(x) if(clock_gettime( CLOCK_REALTIME, &x) == -1 ){ cout<<"clock_gettime error"; };
#endif


using namespace std;

///////////////////////////
// MACROS
///////////////////////////

#define DEFAULT_ITERATIONS 3
// 1.6 GB
#define DEFAULT_BSIZE 1717986918
#define DEFAULT_SSIZE 1
#define DEFAULT_THREADS_NO 1
#define DEFAULT_WARMUP_ITERS 10

#define KB 1024
#define MB (1024*KB)

//I make all bufers aligned
#define ALIGNMENT 16

#define ITERS(count) for(uint64_t i = 0; i < (uint64_t)(count); ++i)

///////////////////////////
// HELPERS
///////////////////////////

void copy_input_buffer(const char* inbuf, char* outbuf, size_t size, size_t bsize, size_t padded_bsize)
{
    while(size)
    {
        size_t current_block_size = min(size, bsize);
        memcpy(outbuf, inbuf, current_block_size);
        size   -= current_block_size;
        inbuf  += current_block_size;
        outbuf += padded_bsize;
    }
}

uint64_t ticks_to_msec(const LARGE_INTEGER& start_ticks, const LARGE_INTEGER& end_ticks, const LARGE_INTEGER& ticksPerSecond)
{
#if defined(WIN32) || defined(_WIN32) || defined(__WIN32__) || defined(_WIN64)
    return (end_ticks.QuadPart - start_ticks.QuadPart)/(ticksPerSecond.QuadPart / 1000);
#else
    return 1000 * ( end_ticks.tv_sec - start_ticks.tv_sec ) + ( end_ticks.tv_nsec - start_ticks.tv_nsec ) / 1000000;
#endif

}

inline size_t round_up(size_t A, size_t B)
{
    // rouds A up to the closest multiply of B
    // note: A has to be in range (-B, INT_MAX-B].
    size_t ret = (A / B) * B; // this rounds towards 0
        if(A > ret) // so we need to round up manually
            ret += B;
    return ret;
}

unsigned getSmallItersCount(const char* inbuf, char* outbuf, size_t size, unsigned threads_no)
{
    // determine the number of small iters, that is how many times do we need to compress a file to get a single timing large enough to be meaningful
    // I assume that 'meaningful' is 'larger than 0.5 s.'
    // I do it by increasing number of small_iters until it reaches 0.25 s. and then calculate the number needed to get 0.5
    // I estimate that the fastest codecs are 3x slower than memcpy
    // It's quite naive, for example doesn't take threading into account
    // Also, it's tweaked for the fastest codecs and too cautious when testing only slow ones; you'll get good results but you'll have to wait for them longer than necessary
    unsigned small_iters;
    LARGE_INTEGER ticksPerSecond;

    InitTimer(ticksPerSecond);

    for(small_iters = 1; small_iters < 8192; small_iters *= 2) // 8192 is a safe default. small_iters can be increased inside the loop 2+epsilon times
    {
        LARGE_INTEGER start_ticks, end_ticks;
        GetTime(start_ticks);
        ITERS(small_iters)
            memcpy(outbuf, inbuf, size);

        GetTime(end_ticks);

        uint64_t msec = ticks_to_msec(start_ticks, end_ticks, ticksPerSecond);
        if(msec > 250)
        {
            uint64_t estimation = small_iters;
            estimation *= threads_no;
            estimation *= 1000;
            estimation /= msec;// enough iterations for 1 second
            estimation /= 2;   // buf I want 1/2 second
            estimation /= 3;   // estimation that codecs are 3x slower than memcpy
            small_iters = max(estimation, (uint64_t)1);
            break;
        }
    }
    if(small_iters < 0)
    {
        cerr << "Something's wrong, can't determine a right number of small iters.\n";
        exit(1);
    }
    return small_iters;
}

void usage()
{
    cerr<<PROGNAME " " PROGVERSION "\n\n";
    cerr<<"usage1: " PROGNAME " help codec_name\n";
    cerr<<"usage2: " PROGNAME " [list of codecs][options] input\n\n";
    cerr<<"list of codecs: a space-separated list of codecs.\n";
    cerr<<"                Each codec is either codec name or a comma separated\n";
    cerr<<"                pair: name,params. See examples below.\n";
    cerr<<"                You can also use the following pseudocodecs:\n";
    cerr<<"                  all:     all supported codecs\n";
    cerr<<"                  fast:    the list of codecs I consider fast\n";
    cerr<<"                  default: the list of codecs used by default\n";
    cerr<<"options:\n";
    cerr<<"  -bX: filesystem block size(default = "<<DEFAULT_BSIZE<<")\n";
    cerr<<"  -c: write output as csv\n";
    cerr<<"  -iX: number of iterations (default = "<<DEFAULT_ITERATIONS<<")\n";
    cerr<<"  -jX: job size (default = determined automatically)\n";
    cerr<<"  -mX: minimal savings, i.e. disk sector size(default = "<<DEFAULT_SSIZE<<")\n";
    cerr<<"  -sX: number of small iterations (default: determined automatically)\n";
    cerr<<"  -tX: number of threads to use(default = "<<DEFAULT_THREADS_NO<<")\n";
    cerr<<"  -v: verify that decoding went fine\n";
    cerr<<"  -wX: warmup iterations(default = "<<DEFAULT_WARMUP_ITERS<<")\n";
    cerr<<endl;
    cerr<<"Available codecs:\n";
    vector<string> codecs_list = pretty_list_of_codecs();
    for(vector<string>::iterator it=codecs_list.begin(); it != codecs_list.end(); ++it)
        cerr<<"  " << *it << endl;
    cerr<<endl;
    cerr<<"Examples:\n";
    cerr<<"  " PROGNAME " help lzo\n";
    cerr<<"  " PROGNAME " file.tar\n";
    cerr<<"  " PROGNAME " -t2 file.tar\n";
    cerr<<"  " PROGNAME " LZ4 zlib,1 -t2 file.tar\n";
    cerr<<"  " PROGNAME " fastlz lzf,ultra -t2 -b131072 -m4096 -i10 file.tar\n";
}

string speed_string(uint64_t bytes, uint64_t msec)
{
    const string units[] = {" B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB"};//OK, uint64_t is not enough to reach YB ;)
    int units_no = sizeof(units) / sizeof(string);
    uint64_t divider = 1;
    for(int i = 0; i < units_no; ++i, divider *= KB)
    {
        uintmax_t size_per_sec = (((uintmax_t)bytes*1000)/msec)/divider;
        if(size_per_sec <= 9999)
            return to_string(size_per_sec) + " "+units[i]+"/s";
    }
    return "fast :)";
}

string efficiency_string(uint64_t saved_bytes, uint64_t msec)
{
    const string units[] = {"e0", "e3", "e6", "e9", "e12", "e15", "e18", "e21", "e24"};//OK, uint64_t is not enough to reach YB ;)
    int units_no = sizeof(units) / sizeof(string);
    uint64_t divider = 1;
    for(int i = 0; i < units_no; ++i, divider *= KB)
    {
        uintmax_t size_per_sec = (((uintmax_t)saved_bytes*1000)/msec)/divider;
        if(size_per_sec <= 9999)
            return to_string(size_per_sec) + units[i];
    }
    return "high :)";
}

///////////////////////////
// OUTPUT PRINTING
///////////////////////////

void printMemcpy(uint64_t milisecs, uint64_t input_size, uint32_t iters, bool csv)
{
    //because of timer inaccuracy some jobs may take 0 ms. And I divide by dmili later...
    milisecs = max(milisecs, (uint64_t)1);
    if(csv)
    {
        cout<<milisecs<<" ms,"<<speed_string(input_size*iters, milisecs)<<","<<input_size<<" B\n";
    }
    else
    {
        streamsize old_width = cout.width(0);
        cout << "memcpy: " << milisecs << " ms, "<<input_size<<" bytes = "<< input_size*iters/milisecs*1000/MB << " MB/s\n";
        cout.width(old_width);
    }
}

void printTime(uint64_t cmili, uint64_t dmili,
    uint64_t input_size, uint64_t compressible_bytes, uint64_t compressed_size, uint32_t iters, bool csv)
{    
    //because of timer inaccuracy some jobs may take 0 ms. And I divide by dmili later...
    dmili = max(dmili, (uint64_t)1);

    if(csv)
    {
        cout<<cmili<<","<<dmili<<","<<compressed_size<<","<<compressible_bytes<<endl;
    }
    else
    {
        ConsoleColour cc(RESULT_COLOUR);
        // I suck at iostream
        char buf [8];
        sprintf(buf, "%.3f", (float)input_size/compressed_size);
        int len = strlen(buf);
        string compression_rate;
        for(int i = 0; i < 6-len; ++i)
            compression_rate += ' ';
        compression_rate += buf;

        streamsize old_width = cout.width(11);
        cout << std::right << compressed_size;
        cout << " (x" << compression_rate << ")      ";
        string decoding_speed_str = compressible_bytes ? speed_string(input_size*iters, dmili) : string("-");
        cout.width(9);
        cout << speed_string(input_size*iters, cmili) << " ";
        cout.width(9);
        cout << decoding_speed_str << "      ";
        cout.width(6);
        cout << efficiency_string(input_size-compressed_size, cmili) << " ";
        cout.width(6);
        cout << efficiency_string(input_size-compressed_size, dmili);

        cout.width(old_width);
    }
}
void printHeaders()
{
    {
        ConsoleColour cc(CODEC_COLOUR);
        cout << "Codec                                   version      args" << endl;
    }
    {
        ConsoleColour cc(RESULT_COLOUR);
        cout << "C.Size      (C.Ratio)        C.Speed   D.Speed      C.Eff. D.Eff." << endl;
    }
}
void printCodec(const Codec& codec, bool csv)
{
    if(csv)
        cout<<codec.name<<","<<codec.version<<","<<codec.args<<",";
    else
    {
        ConsoleColour cc(CODEC_COLOUR);
        cout << codec.introduction()<<endl;
    }
} 


///////////////////////////
// CORRECTNESS VERIFICATION
///////////////////////////

void check_decoding(uint32_t input_crc, char* buf, size_t size, size_t bsize, size_t padded_bsize)
{
    uint32_t output_crc = 0;
    while(size)
    {
        size_t current_block_size = min(size, bsize);
        output_crc = crc(buf, current_block_size, output_crc);
        memset(buf, 0, current_block_size); // clear the output buffer
        buf += padded_bsize;
        size -= current_block_size;
    }
    if(input_crc != output_crc)
    {
        cout << "CRC mismatch!\n";
    }
}

//////////////////////
// ENCODING / DECODING
//////////////////////


// divides data into chunks and precesses them independently
// data format:
// [chunk_size][data...][chunk_size][data...]...
// chunk_size is signed int
// if chunk size > 0 then data is compressed, otherwise - uncompressed
// size of each compressed data block == chunk_size
// if verify == 1, size of uncompressed block == -chunk_size
// if verify == 0, size of uncompressed block == 0, that is - there is none. I save time on memcpy because filesystems don't memcpy anyway
// returns compressed data size, that is - it ignores chunk_size pieces
THREAD_RETURN_TYPE encode(EncodeParams* params)
{
    uintmax_t total = 0; // summing up across small iterations can yield a large number
    uintmax_t successfully_encoded_bytes = 0;

    total = 0;
    
    Scheduler::WorkItem wi;
    while(params->scheduler->getChunk(wi) == 0)
    {
        char* p  = wi.out;
        size_t processed_bytes = 0;
        unsigned i = 0;
        for(; processed_bytes < wi.isize; processed_bytes += params->workbsize, ++i)
        {
            size_t chunk_size = min(wi.isize-processed_bytes, params->bsize);
            size_t ret = params->encoder(wi.in+processed_bytes, chunk_size, p, round_up(chunk_size-params->ssize, params->ssize), params->other);

            bool failed_to_encode = ret == CODING_ERROR;

            size_t real_size = round_up(ret, params->ssize);

            if(!failed_to_encode && params->can_skip)
            {
                failed_to_encode = real_size >= params->bsize || real_size >= chunk_size;
            }
            if(failed_to_encode)
            {
                //FIXME: if not params->can_skip this should return error
                wi.metadata[i].successfully_encoded = false;
                wi.metadata[i].encoded_size = params->bsize;
                wi.metadata[i].decoded_size = params->bsize;
                total += round_up(chunk_size, params->ssize);
                if(params->verify)
                {
                    memcpy(p, wi.in+processed_bytes, chunk_size);
                }
            }
            else
            {
                wi.metadata[i].successfully_encoded = true;
                wi.metadata[i].encoded_size = ret;
                wi.metadata[i].decoded_size = chunk_size;
                total += real_size;
                successfully_encoded_bytes += chunk_size;
                if(params->verify)
                {
                    // zero-out the part of the buffer that doesn't contain encoded data
                    // if a codec cheats by reporting size smaller than what it actually needs,
                    // this will detect it
                    memset(p+ret, 0, params->workbsize-ret);
                }
            }
            p += params->workbsize;
        }
    }
    params->encoded_bytes   = total;
    params->successfully_encoded_bytes = successfully_encoded_bytes;
    return THREAD_RETURN;
}


THREAD_RETURN_TYPE decode(DecodeParams* params)
{
    uintmax_t total_out = 0; // summing up across small iterations can yield a large number
    
    Scheduler::WorkItem wi;
    while(params->scheduler->getChunk(wi) >= 0)
    {
        ptrdiff_t job_out = 0;
        char* ip = wi.in;
        char* op = wi.out;
        unsigned i = 0;
        size_t processed_bytes = 0;
        for(; processed_bytes < wi.isize; processed_bytes += params->workbsize, ++i)
        {
            size_t ichunk_size = wi.metadata[i].encoded_size;
            if(wi.metadata[i].successfully_encoded)
            {
                size_t ochunk_size = wi.metadata[i].decoded_size;
                size_t ret = params->decoder(ip, ichunk_size, op, ochunk_size, params->other);
                
                if(ret == CODING_ERROR)
                {
                    cerr << "Decoding failed!\n";
                    exit(1);
                }
                    
                size_t real_out = (ret / params->ssize) * params->ssize;
                if(real_out != ret)
                    real_out += params->ssize;
                if(job_out + real_out >= wi.isize)
                    job_out += ret;
                else
                    job_out += real_out;
            }
            else
            {
                if(params->verify)
                {
                    memcpy(op, ip, ichunk_size);
                }
                job_out += ichunk_size;
            }
            op += params->workbsize;
            ip += params->workbsize;
        }
        total_out += job_out;
    }
    params->ret = total_out;

    return THREAD_RETURN;
}


////////////////////////////
// THE MAIN TESTING FUNCTION
////////////////////////////

// returns number of small_iters actually used
unsigned test(list<CodecWithParams>& codecs, ifstream& in, unsigned iters, unsigned small_iters, size_t bsize, size_t ssize, bool verify, unsigned warmup_iters, unsigned desired_threads_no, bool csv, size_t job_size)
{

// I run this many times more decoding jobs
#define DECODING_MULTIPLIER 2

    LARGE_INTEGER ticksPerSecond, cstart_ticks, dstart_ticks, cend_ticks, dend_ticks;

    InitTimer(ticksPerSecond);

    in.seekg(0, ios_base::end);
    size_t size = in.tellg();
    in.seekg(0, ios_base::beg);
    
    bsize = min(bsize, size);
    
    if(bsize > DEFAULT_BSIZE)
    {
        cerr << "\nWARNING: You're using a really large block size. Expect troubles.\n\n";
    }
    
    // number of blocks
    size_t blocks = (size / bsize);
    if(blocks*bsize < size)
        ++blocks;

    uint32_t threads_no = min((size_t)desired_threads_no, blocks);
    if(threads_no < desired_threads_no)
    {
        cerr << "\nWARNING: You have specified to use " << desired_threads_no << " threads, but have " << blocks << " block(s).\n";
        cerr << "         Only " << threads_no << " thread(s) will be used.\n";
        cerr << "         To resolve the issue, specify a lower block size.\n\n";
    }
    
    if(size < 16 * MB * threads_no)
        cerr << "\nWARNING: This file is too small, use at least 16 MB per thread.\n\n";
    
    size_t workbuf_block_size = bsize; // I need some space for memcpy
    for(list<CodecWithParams>::iterator it = codecs.begin(); it != codecs.end(); ++it)
        workbuf_block_size = max(workbuf_block_size, it->codec.max_compressed_size(bsize));
    
    workbuf_block_size = round_up(workbuf_block_size + OVERRUN_PAD, ALIGNMENT);

    size_t inbuf_size       = size + OVERRUN_PAD;
    size_t workbuf_size     = workbuf_block_size * blocks;

    // allocate possibly misaligned buffers
    char* _inbuf           = new(nothrow) char     [inbuf_size + ALIGNMENT - 1];// some extra space, so I can align it
    char* _workbuf         = new(nothrow) char     [2 * workbuf_size + ALIGNMENT - 1];// some extra space, so I can align it
    BlockInfo* metadatabuf = new(nothrow) BlockInfo[blocks];

    if (!_inbuf || !_workbuf || !metadatabuf)
    {
        cerr<<"Not enough memory!\n";
        cerr<<"Tried to allocate "<<inbuf_size + ALIGNMENT - 1<<"+"<<2*workbuf_size+ALIGNMENT-1<<"+"<<blocks*sizeof(BlockInfo)<<" bytes.\n";
        exit(1);
    }
    
    // aligned buffers
    char* workbufs[2];
    char* inbuf = (char*)(((uintptr_t)_inbuf   & ~(ALIGNMENT - 1)) ? ((uintptr_t)_inbuf   & ~(ALIGNMENT - 1)) + ALIGNMENT: (uintptr_t)_inbuf   & ~(ALIGNMENT - 1));
    workbufs[0] = (char*)(((uintptr_t)_workbuf & ~(ALIGNMENT - 1)) ? ((uintptr_t)_workbuf & ~(ALIGNMENT - 1)) + ALIGNMENT: (uintptr_t)_workbuf & ~(ALIGNMENT - 1));
    workbufs[1] = workbufs[0] + workbuf_size; // workbuf_size is a multiply of ALIGNMENT, so this one is aligned too

    in.read(inbuf, size);

    if(in.fail())
    {
        cerr<<"Error reading file!\n";
        exit(1);
    }
    
    uint32_t input_crc = crc(inbuf, size, 0);

    ITERS(warmup_iters)
    {
        // I copy some more that I read, but I want just to touch the memory
        memcpy(workbufs[1], inbuf,       min(inbuf_size, workbuf_size));
        memcpy(workbufs[0], workbufs[1], workbuf_size);
    }

    if(!small_iters)
    {
        small_iters = getSmallItersCount(inbuf, workbufs[1], size, threads_no);
    }

    GetTime(cstart_ticks);
    ITERS(small_iters) memcpy(workbufs[1], inbuf, size);
    GetTime(cend_ticks);
    printMemcpy(ticks_to_msec(cstart_ticks, cend_ticks, ticksPerSecond), size, small_iters, csv);
    printHeaders();
    
    THREAD_HANDLE* threads = 0;
    if(threads_no > 1)
    {
        threads = new(nothrow) THREAD_HANDLE[threads_no-1];// The main thread does 1 job, so I need 1 handle less

        if (!threads)
        {
            cerr<<"Not enough memory!\n";
            exit(1);
        }
    }

    // prepare compression parameters
    EncodeParams* params  = new(nothrow) EncodeParams[threads_no];
    DecodeParams* dparams = new(nothrow) DecodeParams[threads_no];
    if(!params || !dparams)
    {
        cerr<<"Out of memory!\n";
        exit(1);
    }

    // the main loop
    for(list<CodecWithParams>::iterator it=codecs.begin(); it != codecs.end(); ++it)
    {
        Codec& codec = it->codec;
        const string& codec_init_params = it->params;

        codec.init(codec_init_params, threads_no, min(size, bsize));
        printCodec(codec, csv);
        if(codec.encoder == 0)
        {
            cerr << "ERROR: " << codec.name << " is just a decoder.\n";
            cerr << "Combine it with some encoder to test it.";
            continue;
        }
        else if(codec.decoder == 0)
        {
            cerr << "ERROR: " << codec.name << " is just an encoder.\n";
            cerr << "Combine it with some decoder to test it.";
            cerr << "If you don't want one, combine it with nop.";
            continue;
        }

        uint32_t fastest_ctime = -1;
        uint32_t fastest_dtime = -1;

        intmax_t complen = -1;
        intmax_t compressible_bytes = -1;// amount of data compressed successfuly. On incompressible files expected to be 0
        
        size_t input_buffer  = -1;
        size_t output_buffer = -1;

        for(unsigned j = 0; j < iters; ++j)
        {
            complen = 0;
            compressible_bytes = 0;
            unsigned i;

            input_buffer = 0;
            
            copy_input_buffer(inbuf, workbufs[input_buffer], size, bsize, workbuf_block_size);
            
            output_buffer = codec.encode_transform_type == Codec::moving ? 1 : 0;
            
            bool last_block_is_full = round_up(size, bsize) == size;
            size_t size_in_blocks = size + (workbuf_block_size - bsize) * (last_block_is_full ? blocks : (blocks - 1));
            
            Scheduler compression_scheduler = Scheduler(workbufs[input_buffer], workbufs[output_buffer], metadatabuf, size_in_blocks, workbuf_block_size, small_iters, job_size);
            // initialize thread data
            void** codec_eparams = codec.eparams();
            for(i = 0; i < threads_no; ++i)
            {
                params[i].encoder   = codec.encoder;
                params[i].other     = codec_eparams ? (void*)&codec_eparams[i] : 0;
                params[i].bsize     = bsize;
                params[i].workbsize = workbuf_block_size;
                params[i].ssize     = ssize;
                params[i].verify    = verify;
                params[i].scheduler = &compression_scheduler;
                params[i].can_skip  = codec.can_be_skipped;
            }

            // ready...set...go!
            GetTime(cstart_ticks);

            // compression
            for(i = 0; i < threads_no-1; ++i)
                if(int ret = pthread_create (threads+i, NULL, (THREAD_RETURN_TYPE(*) (void*))encode, params+i))
                {
                    cerr<<"Can't start a thread: "<<ret<<endl;
                    exit(1);
                }

            // I don't spawn a thread to do the last job, use the current one
            encode(&params[threads_no-1]);

            // wait for other threads to do the job
            for(i = 0; i < threads_no-1; ++i)
            {
                pthread_join (threads[i], NULL);
            }
            
            GetTime(cend_ticks);
            
            for(i = 0; i < threads_no; ++i)
            {
                complen            += params[i].encoded_bytes;
                compressible_bytes += params[i].successfully_encoded_bytes;
            }
            
            complen            /= small_iters;
            compressible_bytes /= small_iters;
            
            if(verify && codec.encode_transform_type == Codec::moving)
                // clear the previous buffer
                memset(workbufs[input_buffer], 0, workbuf_size);
                
            input_buffer  = output_buffer;
            output_buffer = codec.encode_transform_type == Codec::moving ? 1 - input_buffer : input_buffer;

            // decompression is fast. It makes the results more volatile,
            // but it's also cheap to fix by running more of them
            Scheduler decompression_scheduler = Scheduler(workbufs[input_buffer], workbufs[output_buffer], metadatabuf, workbuf_size, workbuf_block_size, small_iters*DECODING_MULTIPLIER, job_size);

            // prepare decompression settings
            void** codec_dparams = codec.dparams();
            for(i = 0; i < threads_no; ++i)
            {
                dparams[i].decoder   = codec.decoder;
                dparams[i].other     = codec_dparams ? (void*)&codec_dparams[i] : 0;
                dparams[i].workbsize = workbuf_block_size;
                dparams[i].ssize     = ssize;
                dparams[i].verify    = verify;      
                dparams[i].scheduler = &decompression_scheduler;
            }
            
            GetTime(dstart_ticks);
            
            // settings for threads that get blocks_per_thread+1 blocks
            for(i = 0; i < threads_no-1; ++i)
            {
                if(int ret = pthread_create (threads+i, NULL, (THREAD_RETURN_TYPE(*) (void*))decode, &dparams[i]))
                {
                    cerr<<"Can't start a thread: "<<ret<<endl;
                    exit(1);
                }
            }
            
            // we're the last thread
            decode(&dparams[threads_no-1]);

            // wait for other threads to do the job
            for(i = 0; i < threads_no-1; ++i)
            {
                pthread_join (threads[i], NULL);
            }
            
            GetTime(dend_ticks);

            uint32_t ctime = ticks_to_msec(cstart_ticks, cend_ticks, ticksPerSecond);
            uint32_t dtime = ticks_to_msec(dstart_ticks, dend_ticks, ticksPerSecond);

            fastest_ctime = min(fastest_ctime, ctime);
            fastest_dtime = min(fastest_dtime, dtime);

            if(!csv)
            {
                printTime(fastest_ctime, fastest_dtime/DECODING_MULTIPLIER, size, compressible_bytes, complen, small_iters, csv);
                cout << '\r';
            }
        }
        if(csv)
            printTime(fastest_ctime, fastest_dtime / 2, size, compressible_bytes, complen, small_iters, csv);
        else
            cout <<endl;
        if(verify)
            check_decoding(input_crc, workbufs[output_buffer], size, bsize, workbuf_block_size);

        codec.cleanup();
    }
    printHeaders();

    delete[] params;
    delete[] dparams;
    if(threads)
        delete[] threads;

    delete[] _inbuf;
    delete[] _workbuf;
    return small_iters;
}

///////////////////////////
// MAIN
///////////////////////////
int main( int argc, char** argv)
{

    // increase priority to the max to reduce variance
    setHighestPriority();

    uint32_t iterations   = DEFAULT_ITERATIONS;
    size_t bsize          = DEFAULT_BSIZE;
    size_t ssize          = DEFAULT_SSIZE;
    bool verify           = false;
    int threads           = DEFAULT_THREADS_NO;
    int warmup_iters      = DEFAULT_WARMUP_ITERS;
    bool csv              = false;// should the output be written as csv or human readably?
    unsigned small_iters  = 0;
    size_t job_size       = 262144;// when compressing small blocks, each work item will contain multiple of them, so the size is no smaller than this

    list<CodecWithParams> codecs;

    //debug code
    /*char*a[] = {"fsbench", "bcl-lzfast", "-w0", "-s1", "-i1", "..\\calgary.tar"};
    argv=a;
    argc=6;*/

    if(argc == 3)
    {
        if(!case_insensitive_compare(argv[1], "help"))
        {
            Codec* c = find_codec(argv[2]);
            if(c)
            {
                cout << c->help();
                exit(0);
            }
            
            cerr<<"Unknown codec: '" <<argv[2]<<"'";
            exit(1);
        }
    }

    while(argc > 2)
    {
        if(argv[1][0] == '-')
        {
            switch (argv[1][1])
            {
            case 'b':
                bsize = atoi(argv[1]+2);
                if(bsize <= 0)
                {
                    cerr<<"block size must be > 0.\n";
                    exit(1);
                }
                break;
            case 'c':
                csv = true;
                break;
            case 'i':
                iterations = atoi(argv[1]+2);
                if(iterations <= 0)
                {
                    cerr<<"number of iterations must be > 0.\n";
                    exit(1);
                }
                break;
            case 'j':
                job_size = atoi(argv[1]+2);
                if(job_size <= 0)
                {
                    cerr<<"job size must be > 0.\n";
                    exit(1);
                }
                break;
            case 'm':
                ssize = atoi(argv[1]+2);
                if(ssize <= 0)
                {
                    cerr<<"minimum savings must be > 0.\n";
                    exit(1);
                }
                break;
            case 's':
                small_iters = atoi(argv[1]+2);
                if(small_iters <= 0)
                {
                    cerr<<"number of small iterations must be > 0.\n";
                    exit(1);
                }
                break;
            case 't':
                threads = atoi(argv[1]+2);
                if(threads <= 0)
                {
                    cerr<<"number of threads must be > 0.\n";
                    exit(1);
                }
                break;
            case 'v':
                verify = true;
                break;
            case 'w':
                warmup_iters = atoi(argv[1]+2);
                if(warmup_iters < 0)
                {
                    cerr<<"warmup_iters of threads must be >= 0.\n";
                    exit(1);
                }
                break;
            default:
                cerr<<"unknown option: '"<<argv[1]<<"'\n";
                exit(1);
            }
        }
        else
        {
            // it may be "codec" or "codec,params" or "pseudocodec".
            // I'll check the first option and if it's not the case,
            // try splitting
            if(case_insensitive_compare(argv[1], "all") == 0)
            {
                codecs.insert(codecs.end(), ALL_CODECS.begin(), ALL_CODECS.end());
            }
            else if(case_insensitive_compare(argv[1], "default") == 0)
            {
                codecs.insert(codecs.end(), DEFAULT_CODECS.begin(), DEFAULT_CODECS.end());
            }
            else if(case_insensitive_compare(argv[1], "fast") == 0)
            {
                codecs.insert(codecs.end(), FAST_CODECS.begin(), FAST_CODECS.end());
            }
            else
            {
                string codec = argv[1];
                string params = "";
                size_t position = codec.find(',');
                if(position != string::npos)
                {
                    params = codec.substr(position + 1);
                    codec = codec.substr(0, position);
                }
                Codec* found = find_codec(codec);
                if(found)
                {
                    codecs.push_back(CodecWithParams(*found, params));
                }
                else
                {
                    cerr<<"Unknown codec: '"<<codec<<"'.";
                    exit(1);
                }
            }
        }
        argv++;
        argc--;
    }

    if (argc < 2) 
    {
        usage();
        exit(1);
    }

    ifstream in (argv[1], ios::in|ios::binary);
    if(in.fail())
    {
        cerr<<"Failed to open the input file.\n";
        exit(1);
    }

    if(codecs.empty())
        codecs = DEFAULT_CODECS;

    try
    {
        small_iters = test(codecs, in, iterations, small_iters, bsize, ssize, verify, warmup_iters, threads, csv, job_size);

        cout<<"done... ("<<iterations<<'x'<<small_iters<<" iteration(s)).\n";
    }
    catch(const Codec::InvalidParams &p)
    {
        cerr << "Invalid params!\n";
        cerr << p.message;
        exit(1);
    }
}