/*
    Implementations of the Codec interface
    
    Written by m^2.
    You can consider the mods to be public domain if you care to separate them.
    If your country doesn't recognize author's right to relieve themselves of copyright,
    you can use it under the terms of WTFPL version 2.0 or later.

*/

#include "codecs.hpp"
#include "abstractCodecs.hpp"
#include "tools.hpp"

#include <algorithm>
#include <climits>
#include <cstring>
#include <iostream>
#include <map>
#include <memory.h>

#if __cplusplus >= 201103L // C++ 2011
    #include <cstdint>
#else
extern "C"
{
    #include <stdint.h>
}
#endif // C++ 2011

using namespace std;

//////////////////////////////////////////////////////
// Codecs that use some of the abstract classes
// Here they have functions used by their constructors
//////////////////////////////////////////////////////

#ifdef FSBENCH_USE_7Z
#include "7z/7z.hpp"
#endif//FSBENCH_USE_7Z

#ifdef FSBENCH_USE_BCL
#include "bcl.hpp"
#endif//FSBENCH_USE_BCL

#ifdef FSBENCH_USE_BLOSC
extern "C"
{
    #include "blosc/blosclz.h"
}

size_t blosc_c(char*in,size_t isize,char*out,size_t osize,void*mode){int ret = blosclz_compress(*(intptr_t*)mode, in, isize, out, osize); return ret > 0 ? ret : CODING_ERROR;}
size_t blosc_d(char*in,size_t isize,char*out,size_t osize,void*_)   {return blosclz_decompress(in, isize, out, osize);}
#endif//FSBENCH_USE_BLOSC
#ifdef FSBENCH_USE_BZ2
extern "C"
{
    #include "bzlib.h"
}

size_t bz2_c(char* in, size_t isize, char* out, size_t osize, void* mode)
{
    unsigned int _osize = osize;
    int ret = BZ2_bzBuffToBuffCompress(out, &_osize, in, isize, *(intptr_t*)mode, 0, 0); 
    return ret == BZ_OK ? _osize : CODING_ERROR;
}
size_t bz2_d(char* in, size_t isize, char* out, size_t osize, void* _)
{

    unsigned int _osize = osize;
    int ret = BZ2_bzBuffToBuffDecompress(out, &_osize, in, isize, 0, 0); 
    return ret == BZ_OK ? _osize : CODING_ERROR;
}
#endif//FSBENCH_USE_BZ2
#ifdef FSBENCH_USE_ECRYPT
#include "ecrypt/fsbench_ecrypt.hpp"
#endif//FSBENCH_USE_ECRYPT

#ifdef FSBENCH_USE_CITYHASH
#include "CityHash/city.h"
void CityHash64(char* in, size_t isize, char* out)
{
    *(uint64_t*) out = CityHash64((const char*)in, isize);
}
void CityHash128(char* in, size_t isize, char* out)
{
    uint128 ret = CityHash128((const char*)in, isize);
    ((uint64_t*)out)[0] = Uint128Low64(ret);
    ((uint64_t*)out)[1] = Uint128High64(ret);
}
#endif//FSBENCH_USE_CITYHASH

#ifdef FSBENCH_USE_CRAPWOW
extern "C"
{
    #include "CrapWow.h"
}
void CrapWow(char* in, size_t isize, char* out)
{
    *(uint32_t*) out = CrapWow((const uint8_t*)in, (uint32_t)isize, (uint32_t)0);
}
#endif//FSBENCH_USE_CRAPWOW

#ifdef FSBENCH_USE_CRYPTOPP
#include "cryptopp/cryptopp.hpp"
#endif//FSBENCH_USE_CRYPTOPP

#ifdef FSBENCH_USE_DOBOZ
#include "Doboz/Compressor.h"
#include "Doboz/Decompressor.h"
size_t Doboz_c(char*in,size_t isize,char*out,size_t osize,void*_){doboz::Compressor c;size_t ret;return c.compress(in,isize,out,c.getMaxCompressedSize(isize),ret) == doboz::RESULT_OK ? ret : CODING_ERROR;}
size_t Doboz_d(char*in,size_t isize,char*out,size_t osize,void*_){doboz::Decompressor d;return d.decompress(in,isize,out,osize) == doboz::RESULT_OK ? osize : CODING_ERROR;}
#endif//FSBENCH_USE_DOBOZ
#ifdef FSBENCH_USE_FASTLZ
extern "C"
{
    #include "fastlz.h"
}
size_t fastlz1_c(char*in,size_t isize,char*out,size_t osize,void*_){return fastlz_compress_level(1, in, isize, out);}
size_t fastlz2_c(char*in,size_t isize,char*out,size_t osize,void*_){return fastlz_compress_level(2, in, isize, out);}
size_t fastlz_d (char*in,size_t isize,char*out,size_t osize,void*_){return fastlz_decompress(in, isize, out, osize);}
size_t fastlz_m (size_t input_size)                                 {return max((size_t)66, input_size + input_size/20 + 1);}

#endif//FSBENCH_USE_FASTLZ

#ifdef FSBENCH_USE_HALIBUT

#define HALIBUT_BUF_SIZE (1 * 1024 * 1024)
extern "C"
{
#include "halibut-deflate.h"
}
size_t halibut_c(char* in, size_t isize, char* out, size_t osize, void* _)
{
    size_t total_out = 0;
    deflate_compress_ctx* ctx = deflate_compress_new(DEFLATE_TYPE_BARE);
    do
    {
        void* tmp_buf = 0;
        int received_size = 0;
        int chunk_size = min((size_t)HALIBUT_BUF_SIZE, isize);
        int flush = isize > HALIBUT_BUF_SIZE ? DEFLATE_NO_FLUSH : DEFLATE_END_OF_DATA;
        deflate_compress_data(ctx, in, chunk_size, flush, &tmp_buf, &received_size);
        if(received_size)
        {
            if((size_t)received_size > osize)
            {
                total_out = CODING_ERROR;
                break;
            }
            memcpy(out, tmp_buf, received_size);

            total_out += received_size;
            out       += received_size;
            osize     -= received_size;

            free(tmp_buf);
        }
        in += chunk_size;
        isize -= chunk_size;

    } while(isize > 0);
    deflate_compress_free(ctx);
    return total_out;
}
size_t halibut_d(char* in, size_t isize, char* out, size_t osize, void* _)
{
    size_t total_out = 0;
    deflate_decompress_ctx* ctx = deflate_decompress_new(DEFLATE_TYPE_BARE);
    do
    {
        void* tmp_buf = 0;
        int received_size = 0;

        int chunk_size = min((size_t)HALIBUT_BUF_SIZE, isize);
        int ret = deflate_decompress_data(ctx, in, chunk_size, &tmp_buf, &received_size);
        if(ret != DEFLATE_NO_ERR)
        {
            total_out = CODING_ERROR;
            break;
        }
        if(received_size)
        {
            if((size_t)received_size > osize)
            {
                total_out = CODING_ERROR;
                break;
            }
            memcpy(out, tmp_buf, received_size);

            total_out += received_size;
            out += received_size;
            osize -= received_size;

            free(tmp_buf);
        }
        in += chunk_size;
        isize -= chunk_size;

    } while(isize > 0);
    deflate_decompress_free(ctx);
    return total_out;
}

#endif//FSBENCH_USE_HALIBUT
#ifdef FSBENCH_USE_LODEPNG
#include "lodepng.h"
size_t lodepng_c(char* in, size_t isize, char* out, size_t osize, void* mode)
{
    LodePNGCompressSettings settings;
    settings.btype      = 2;
    settings.use_lz77   = 1;
    settings.windowsize = *(intptr_t*)mode;

    unsigned char* tmp = 0;
    size_t _osize = 0;
    unsigned ret = lodepng_deflate(&tmp, &_osize, (unsigned char*)in, isize, &settings);
    if(ret != 0)
        return CODING_ERROR;
    memcpy(out, tmp, _osize);
    free(tmp);
    return _osize;
}
size_t lodepng_d(char* in, size_t isize, char* out, size_t osize, void* _)
{
    unsigned char* tmp = 0;
    size_t _osize = 0;
    unsigned ret = lodepng_inflate(&tmp, &_osize, (unsigned char*)in, isize);
    if(ret != 0)
    return CODING_ERROR;
    memcpy(out, tmp, _osize);
    free(tmp);
    return _osize;
}
#endif//FSBENCH_USE_LODEPNG

#ifdef FSBENCH_USE_LZ4
extern "C"
{
    #include "lz4.h"
    #include "lz4hc.h"
}
size_t LZ4_c(char*in,size_t isize,char*out,size_t osize,void*_)        {return LZ4_compress(in, out, isize);}
size_t LZ4_17_c(char*in,size_t isize,char*out,size_t osize,void*_)     {return LZ4_compress_17(in, out, isize);}
size_t LZ4hc_c(char*in,size_t isize,char*out,size_t osize,void*_)      {return LZ4_compressHC(in, out, isize);}
size_t LZ4_d(char*in,size_t isize,char*out,size_t osize,void*_)        {return LZ4_uncompress(in, out, osize) > 0 ? osize : CODING_ERROR;}
size_t LZ4_m(size_t input_size)                                         {return LZ4_compressBound(input_size);}

#endif//FSBENCH_USE_LZ4

#ifdef FSBENCH_USE_LZF

extern "C"
{
    #include "lzf.h"
}
size_t LZF_ultra_c(char*in,size_t isize,char*out,size_t osize,void*_){return lzf_compress_ultra(in, isize, out, osize);}
size_t LZF_very_c(char*in,size_t isize,char*out,size_t osize,void*_) {return lzf_compress_very(in, isize, out, osize);}
size_t LZF_c(char*in,size_t isize,char*out,size_t osize,void*_)      {return lzf_compress(in, isize, out, osize);}
size_t LZF_d(char*in,size_t isize,char*out,size_t osize,void*_)      {return lzf_decompress(in, isize, out, osize);}

#endif//FSBENCH_USE_LZF
#ifdef FSBENCH_USE_LZFX

extern "C"
{
    #include "lzfx.h"
}
size_t LZFX_c(char*in,size_t isize,char*out,size_t osize,void*_){unsigned int ret=osize;return lzfx_compress  (in, isize, out, &ret)>=0 ? ret : CODING_ERROR;}
size_t LZFX_d(char*in,size_t isize,char*out,size_t osize,void*_){unsigned int ret=osize;return lzfx_decompress(in, isize, out, &ret)>=0 ? ret : CODING_ERROR;}

#endif//FSBENCH_USE_LZFX
#ifdef FSBENCH_USE_LZG

extern "C"
{
    #include "liblzg/include/lzg.h"
}
size_t LZG_c(char* in, size_t isize, char* out, size_t osize, void* mode)      
{
    lzg_encoder_config_t cfg;
    cfg.level       = *(intptr_t*)mode;
    cfg.fast        = LZG_TRUE;
    cfg.progressfun = NULL;
    cfg.userdata    = NULL;
    return LZG_Encode((const unsigned char*)in, isize, (unsigned char*)out, LZG_MaxEncodedSize(isize), &cfg);
}
size_t LZG_d(char*in,size_t isize,char*out,size_t osize,void*_){return LZG_Decode((const unsigned char*)in, isize, (unsigned char*)out, osize);}
size_t LZG_m(size_t input_size)                                 {return LZG_MaxEncodedSize(input_size);}

#endif//FSBENCH_USE_LZG
#ifdef FSBENCH_USE_LZMAT

extern "C"
{
    #include "lzmat.h"
}
size_t lzmat_c(char*in,size_t isize,char*out,size_t osize,void*_){MP_U32 retsize=osize; int ret = lzmat_encode((MP_U8*)out, &retsize, (MP_U8*)in, isize); return ret == LZMAT_STATUS_OK ? retsize : CODING_ERROR;}
size_t lzmat_d(char*in,size_t isize,char*out,size_t osize,void*_){MP_U32 ret=osize; lzmat_decode((MP_U8*)out, &ret, (MP_U8*)in, isize);return ret;}

#endif//FSBENCH_USE_LZMAT
#ifdef FSBENCH_USE_LZX_COMPRESS

extern "C"
{
    #include "lzx_compress/lzx_compress.h"
}
int LZX_get_bytes(void *arg, int n, void *buf)
{
    pair<char*,int>* data = (pair<char*,int>*)arg;
    int bytes_to_copy = min(n, data->second);
    memcpy(buf, data->first, bytes_to_copy);
    data->first  += bytes_to_copy;
    data->second -= bytes_to_copy;
    return bytes_to_copy;
}
int LZX_at_eof(void *arg)
{
    pair<char*,int>* data = (pair<char*,int>*)arg;
    return data->second;
}
int LZX_put_bytes(void *arg, int n, void *buf)
{
    pair<char*,size_t>* data = (pair<char*,size_t>*)arg;
    size_t bytes_to_copy = min((size_t)n, data->second);
    memcpy(data->first, buf, bytes_to_copy);
    data->first  += bytes_to_copy;
    data->second -= bytes_to_copy;
    return bytes_to_copy;
}

size_t LZX_c(char* in, size_t isize, char* out, size_t osize, void* window_size)
{
    lzx_data *lzxd;
    lzx_results lzxr;
    
    pair<char*,size_t> get_data;
    pair<char*,size_t> put_data;

    get_data.first  = in;
    get_data.second = isize;
    put_data.first  = out;
    put_data.second = osize;
    
    lzx_init(&lzxd, *(intptr_t*)window_size, LZX_get_bytes, &get_data, LZX_at_eof, 
             LZX_put_bytes, &put_data, NULL, NULL);
    
    size_t block_size = 1 << *(intptr_t*)window_size;
    
    while(get_data.second > 0)
    {
        lzx_reset(lzxd);
        
        size_t used_block_size = min(block_size, get_data.second);
    
        lzx_compress_block(lzxd, used_block_size, 1);
    }
    lzx_finish(lzxd, &lzxr);
    
    return lzxr.len_compressed_output;
}
size_t LZX_d(char* in, size_t isize, char* out, size_t osize, void* memory)
{
    return osize;
}
#endif// FSBENCH_USE_LZX_COMPRESS

#ifdef FSBENCH_USE_MINIZ

#include "miniz.hpp"

#endif//FSBENCH_USE_MINIZ

#ifdef FSBENCH_USE_MURMUR
#include "MurmurHash3.h"
void murmur_x86_32(char* in, size_t isize, char* out)
{
    MurmurHash3_x86_32(in, isize, 0, out);
}
void murmur_x86_128(char* in, size_t isize, char* out)
{
    MurmurHash3_x86_128(in, isize, 0, out);
}
void murmur_x64_128(char* in, size_t isize, char* out)
{
    MurmurHash3_x64_128(in, isize, 0, out);
}
#endif//FSBENCH_USE_MURMUR

#ifdef FSBENCH_USE_QUICKLZZIP

extern "C"
{
#include "quicklzzip.h"
}
size_t qlzzip_c(char* in, size_t isize, char* out, size_t osize, void* _)
{
    unsigned out_bits = qlz_deflate((unsigned char*)in, (unsigned char*)out, 0, (unsigned)isize, 1);
    // round up
    if(out_bits & 7)
        out_bits += 8;
    return out_bits / 8;
}

#endif//FSBENCH_USE_QUICKLZZIP

#ifdef FSBENCH_USE_RLE64

#include "RLE64.hpp"

size_t RLE64_c(char*in,size_t isize,char*out,size_t osize,void*_){return RLE64_Compress64(in, out, isize);}
size_t RLE64_d(char*in,size_t isize,char*out,size_t osize,void*_){return RLE64_Uncompress64(in, out, isize);}
size_t RLE32_c(char*in,size_t isize,char*out,size_t osize,void*_){return RLE64_Compress32(in, out, isize);}
size_t RLE32_d(char*in,size_t isize,char*out,size_t osize,void*_){return RLE64_Uncompress32(in, out, isize);}
size_t RLE16_c(char*in,size_t isize,char*out,size_t osize,void*_){return RLE64_Compress16(in, out, isize);}
size_t RLE16_d(char*in,size_t isize,char*out,size_t osize,void*_){return RLE64_Uncompress16(in, out, isize);}
size_t RLE8_c(char*in,size_t isize,char*out,size_t osize,void*_) {return RLE64_Compress8(in, out, isize);}
size_t RLE8_d(char*in,size_t isize,char*out,size_t osize,void*_) {return RLE64_Uncompress8(in, out, isize);}

#endif// FSBENCH_USE_RLE64

#ifdef FSBENCH_USE_SANMAYCE_FNV
extern "C"
{
    #include "sanmayce.h"
}
void fnv1_tesla(char* in, size_t isize, char* out)
{
    *(uint64_t*) out = FNV1A_Hash_Tesla(in, isize);
}
void fnv1_tesla3(char* in, size_t isize, char* out)
{
    *(uint64_t*) out = FNV1A_Hash_Tesla3(in, isize);
}
void fnv1_yoshimura(char* in, size_t isize, char* out)
{
    *(uint32_t*) out = FNV1A_Hash_Yoshimura(in, isize);
}
void fnv1_jesteress(char* in, size_t isize, char* out)
{
    *(uint32_t*) out = FNV1A_Hash_Jesteress(in, isize);
}
void fnv1_meiyan(char* in, size_t isize, char* out)
{
    *(uint32_t*) out = FNV1A_Hash_Meiyan(in, isize);
}
void fnv1_mantis(char* in, size_t isize, char* out)
{
    *(uint32_t*) out = FNV1A_Hash_Mantis(in, isize);
}
#endif//FSBENCH_USE_SANMAYCE_FNV

#ifdef FSBENCH_USE_SCZ

#include "scz.hpp"

#endif// FSBENCH_USE_SCZ

#ifdef FSBENCH_USE_SHRINKER

extern "C"
{
    #include "Shrinker.h"
}

#define SHRINKER_MIN_SIZE (32)
#define SHRINKER_MAX_SIZE ((1 << 27) - 1)
inline int32_t get_next_shrinker_chunk_size(size_t size_left)
{
    if(size_left >= SHRINKER_MAX_SIZE + SHRINKER_MIN_SIZE)
        return SHRINKER_MAX_SIZE;
    if(size_left > SHRINKER_MAX_SIZE)
        return size_left / 2; // when there's just slightly over SHRINKER_MAX_SIZE left, the last 2 chunks are roughly equal
    return size_left;
}

size_t Shrinker_c(char* in, size_t isize, char* out, size_t osize, void* _)
{
    size_t total_out = 0;
    
    while(isize > 0)
    {
        size_t chunk_size = get_next_shrinker_chunk_size(isize);
        int ret = shrinker_compress(in, out+sizeof(int32_t), chunk_size);
        if(ret == -1)
            return CODING_ERROR;
        *(int32_t*) out = ret;
        out += sizeof(uint32_t);
        
        total_out += ret + sizeof(int32_t); // I'm not sure if it's right, but in the output size I include my metadata size
                                            // Either way, it's required by the Codec interface
        in        += chunk_size;
        isize     -= chunk_size;
        out       += ret;
    }
    return total_out;
}
size_t Shrinker_d(char* in, size_t isize, char* out, size_t osize, void* _)
{
    size_t osize_left = osize;
    
    while(osize_left > 0)
    {
        size_t chunk_size = get_next_shrinker_chunk_size(osize_left);
        int32_t compressed_size = *(int32_t*)in;
        in += sizeof(int32_t);

        int ret = shrinker_decompress(in, out, chunk_size);
        if(ret == -1)
            return CODING_ERROR;
        in         += compressed_size;
        out        += chunk_size;
        osize_left -= chunk_size;
    }
    return osize;
}
size_t Shrinker_m(size_t input_size)
{
    return input_size + (input_size / SHRINKER_MAX_SIZE + 1) * sizeof(int); //overhead of 1 int per SHRINKER_MAX_SIZE block in my code
}
#endif// FSBENCH_USE_SHRINKER
#ifdef FSBENCH_USE_SNAPPY

#include "snappy/snappy.h"

size_t snappy_c(char*in,size_t isize,char*out,size_t osize,void*_){size_t ret;snappy::RawCompress(in, isize, out, &ret);return ret;}
size_t snappy_d(char*in,size_t isize,char*out,size_t osize,void*_)
{
    if(!snappy::RawUncompress(in, isize, out))
        return CODING_ERROR;
    return osize;
}
size_t snappy_m(size_t input_size){return snappy::MaxCompressedLength(input_size);}

#endif// FSBENCH_USE_SNAPPY

#ifdef FSBENCH_USE_SPOOKY
#include "spooky.h"
void spooky(char* in, size_t isize, char* out)
{
    uint64_t d1=0, d2=0;
    SpookyHash::Hash128(in, isize, &d1, &d2);
    memcpy(out,              &d1, sizeof(d1));
    memcpy(out + sizeof(d1), &d2, sizeof(d2));
    
}
#endif//FSBENCH_USE_SPOOKY

#ifdef FSBENCH_USE_TINF
#include "tinf/tinf.hpp"
#endif//FSBENCH_USE_TINF

#ifdef FSBENCH_USE_XXHASH
extern "C"
{
    #include "xxhash.h"
    #include "xxhash256.h"
}
void xxhash_fast(char* in, size_t isize, char* out)
{
    *(unsigned int*) out = XXH_fast32(in, isize, 0);
}
void xxhash_strong(char* in, size_t isize, char* out)
{
    *(unsigned int*) out = XXH_strong32(in, isize, 0);
}
void xxhash_256(char* in, size_t isize, char* out)
{
    uint64_t* seed_out = (uint64_t*)out;
    seed_out[0] = seed_out[1] = seed_out[2] = seed_out[3] = 0;

    XXH_256(in, isize, seed_out);
}
#endif//FSBENCH_USE_XXHASH

#ifdef FSBENCH_USE_ZFS

extern "C"
{
    #include "zfs/zfs.h"
}
size_t LZJB_c(char*in,size_t isize,char*out,size_t osize,void*_){size_t ret = lzjb_compress2010((uint8_t*)in, (uint8_t*)out, isize, osize, 0); return ret == isize ? CODING_ERROR : ret;}
size_t LZJB_d(char*in,size_t isize,char*out,size_t osize,void*_){return lzjb_decompress2010((uint8_t*)in, (uint8_t*)out, isize, osize, 0);}
void fletcher2(char* in, size_t isize, char* out)
{
    fletcher_2_native(in, isize, (uint64_t*)out);
}
void fletcher4(char* in, size_t isize, char* out)
{
    fletcher_4_native(in, isize, (uint64_t*)out);
}
#endif//FSBENCH_USE_ZFS

#ifdef FSBENCH_USE_ZLIB
extern "C"
{
    #include "zlib/zlib.h"
}
size_t zlib_c(char* in, size_t isize, char* out, size_t osize, void* mode)
{
    z_stream stream;
    stream.zalloc = Z_NULL;
    stream.zfree  = Z_NULL;
    stream.opaque = NULL;
    
    int ret = deflateInit2(&stream, *(intptr_t*)mode, Z_DEFLATED, -15, 9, Z_DEFAULT_STRATEGY);
    if(ret != Z_OK)
        return CODING_ERROR;
    stream.next_in   = (Bytef*)in;
    stream.avail_in  = isize;
    stream.next_out  = (Bytef*)out;
    stream.avail_out = osize;
    ret = deflate(&stream, Z_FINISH);
    if(ret != Z_STREAM_END)
        return CODING_ERROR;
    ret = deflateEnd(&stream);
    if(ret != Z_OK)
        return CODING_ERROR;
    return stream.total_out;
}
size_t zlib_d(char* in, size_t isize, char* out, size_t osize, void* _)
{
    z_stream stream;
    stream.zalloc = Z_NULL;
    stream.zfree  = Z_NULL;
    stream.opaque = NULL;
    
    int ret = inflateInit2(&stream, -15);
    if(ret != Z_OK)
        return CODING_ERROR;
    stream.next_in   = (Bytef*)in;
    stream.avail_in  = isize;
    stream.next_out  = (Bytef*)out;
    stream.avail_out = osize;
    ret = inflate(&stream, Z_FINISH);
    if(ret != Z_STREAM_END)
        return CODING_ERROR;
    ret = inflateEnd(&stream);
    if(ret != Z_OK)
        return CODING_ERROR;
    return stream.total_out;
}
#endif //FSBENCH_USE_ZLIB

// a pseudocodec that does nothing.
size_t nop_c(char*in,size_t isize,char*out,size_t osize,void*_){return 1;}
size_t nop_d(char*in,size_t isize,char*out,size_t osize,void*_){return osize;}


#ifdef FSBENCH_USE_BLZ

extern "C"
{
    #include "brieflz/brieflz.h"
}

struct BriefLZ : Codec
{
private:

    static size_t compressor(char* in, size_t isize, char* out, size_t osize, void* work)
    {
        return blz_pack(in, out, isize, *(void**)work);
    }
    static size_t decompressor(char* in, size_t isize, char* out, size_t osize, void* work)
    {
        return blz_depack(in, out, osize);
    }
    char* workmem;
    
    static size_t max_compressed_size(size_t input_size)
    {
        return blz_max_packed_size(input_size);
    }

public:

    BriefLZ() :
        Codec("BriefLZ", _BLZ_VERSION, compressor, decompressor, max_compressed_size),
        workmem(0),
        params(0)
    {}

    virtual void init(const string& args, unsigned threads_no, size_t isize, bool init_compressor, bool init_decompressor)
    {
        if(init_compressor)
        {
            int piece_size = blz_workmem_size(isize);
            this->workmem = new char[threads_no*piece_size];
            this->params = new void*[threads_no];
            for(unsigned i=0; i<threads_no; ++i)
                this->params[i] = workmem + i*piece_size;
        }
    }
    virtual void cleanup()
    {
        if(this->params)
        {
            delete[] this->workmem;
            delete[] this->params;
            this->workmem = 0;
            this->params  = 0;
        }
    }
    string help() const
    {
        return this->name + " " + this->version + "\nTakes no parameters.\n";
    }
    
    void** eparams()
    {
        return params;
    }

protected:
    void** params;
};

#endif// FSBENCH_USE_BLZ

#ifdef FSBENCH_USE_LZHAM

#define LZHAM_NO_ZLIB_COMPATIBLE_NAMES
#include "lzham/include/lzham.h"

struct LZHAM : Codec
{
private:

    struct DecompressionState
    {
        lzham_decompress_state_ptr decompression_state;
        lzham_decompress_params* decompression_params;
    };

    static const string default_mode;
    unsigned threads_no;

    const string& default_args(const string& args)
    {
        return args == "" ? default_mode : args;
    }
    static size_t compressor(char* in, size_t _isize, char* out, size_t _osize, void* _data)
    {
        lzham_compress_state_ptr data = *(lzham_compress_state_ptr*)_data;
        lzham_compress_reinit(data);
        size_t total_in = 0;
        size_t total_out = 0;
        lzham_compress_status_t status;
        do
        {
            size_t isize = _isize - total_in;
            size_t osize = _osize - total_out;
            status = lzham_compress2(
                  data,
                  (const lzham_uint8*)(in  + total_in),  &isize,
                  (lzham_uint8*)      (out + total_out), &osize,
                  LZHAM_FINISH);
            total_in  += isize;
            total_out += osize;
        } while (status == LZHAM_COMP_STATUS_NOT_FINISHED && _osize > total_out && _isize > total_in);

        return status == LZHAM_COMP_STATUS_SUCCESS ? total_out : CODING_ERROR;
    }
    static size_t decompressor(char* in, size_t _isize, char* out, size_t _osize, void* _data)
    {
        lzham_decompress_state_ptr state = (*(DecompressionState**)_data)->decompression_state;
        lzham_decompress_params* params =  (*(DecompressionState**)_data)->decompression_params;

        lzham_decompress_reinit(state, params);

        size_t isize = _isize;
        size_t osize = _osize;

        lzham_decompress_status_t status = lzham_decompress(state, (const lzham_uint8*)in, &isize, (lzham_uint8*)out, &osize, true);
        if (status != LZHAM_DECOMP_STATUS_SUCCESS)
        {
            cerr<<"decompression error:"<<status<<endl;
            return CODING_ERROR;
        }

        return osize;
    }

public:

    LZHAM() :
        Codec("LZHAM", _LZHAM_VERSION, compressor, decompressor, no_blowup),
        compressor_params(0),
        decompressor_params(0)
    {}
    virtual void init(const string& args, unsigned threads_no, size_t isize, bool init_compressor, bool init_decompressor)
    {        
        this->args = default_args(args);
        uintptr_t mode;
        try
        {
            from_string(this->args, mode);
        }
        catch(const bad_cast &bc)
        {
            throw InvalidParams(help());
        }

        if(init_compressor)
        {
            if(mode < 0 || mode > (int)LZHAM_TOTAL_COMP_LEVELS)
                throw InvalidParams(help());

            lzham_compress_params comp_params;
            memset(&comp_params, 0, sizeof(comp_params));
            comp_params.m_struct_size        = sizeof(comp_params);
            comp_params.m_dict_size_log2     = 17;
            comp_params.m_level              = (lzham_compress_level)mode;
            comp_params.m_max_helper_threads = 0;
            comp_params.m_compress_flags     = 0;

            this->compressor_params = new void*[threads_no];
            for(unsigned i=0; i<threads_no; ++i)
            {
                lzham_compress_state_ptr state = lzham_compress_init(&comp_params);
                this->compressor_params[i] = (void*)state;
            }
        }
        if(init_decompressor)
        {
            this->decompressor_params = new void*[threads_no];
            for(unsigned i=0; i<threads_no; ++i)
            {
                lzham_decompress_params* decomp_params = new lzham_decompress_params();
                memset(decomp_params, 0, sizeof(lzham_decompress_params));
                decomp_params->m_struct_size = sizeof(lzham_decompress_params);
                decomp_params->m_dict_size_log2 = 17;
                decomp_params->m_decompress_flags = LZHAM_DECOMP_FLAG_OUTPUT_UNBUFFERED;

                DecompressionState* state = new DecompressionState();

                state->decompression_state = lzham_decompress_init(decomp_params);
                state->decompression_params = decomp_params;
                this->decompressor_params[i] = (void*)state;
            }
        }
        this->threads_no = threads_no;
    }
    virtual void cleanup()
    {
        if(this->compressor_params)
        {
            for(unsigned i=0; i<this->threads_no; ++i)
            {
                lzham_compress_state_ptr state = (lzham_compress_state_ptr) this->compressor_params[i];
                lzham_compress_deinit(state);
            }
            delete[] this->compressor_params;
            this->compressor_params = 0;
        }
        if(this->decompressor_params)
        {
            for(unsigned i=0; i<this->threads_no; ++i)
            {
                DecompressionState* state = (DecompressionState*) this->decompressor_params[i];
                lzham_decompress_deinit(state->decompression_state);
                delete state->decompression_params;
                delete state;
            }
            delete[] this->decompressor_params;
            this->decompressor_params = 0;
        }
    }

    virtual string help() const
    {
        return "LZHAM " + this->version + "\nAvaivable parameters: \n Numbers from 0 to 4\n";
    }

    void** eparams()
    {
        return this->compressor_params;
    }
    void** dparams()
    {
        return this->decompressor_params;
    }

protected:
    void** compressor_params;
    void** decompressor_params;
};
const string LZHAM::default_mode("4");

#endif

#ifdef FSBENCH_USE_LZO
extern "C"
{
    #include "lzo/lzo1.h"
    #include "lzo/lzo1a.h"
    #include "lzo/lzo1b.h"
    #include "lzo/lzo1c.h"
    #include "lzo/lzo1f.h"
    #include "lzo/lzo1x.h"
    #include "lzo/lzo1y.h"
    #include "lzo/lzo1z.h"
    #include "lzo/lzo2a.h"
}

struct LZO : Codec
{
private:

    typedef int
        (*lzo_compress)( const lzo_bytep src, lzo_uint  src_len,
                                    lzo_bytep dst, lzo_uintp dst_len,
                                    lzo_voidp wrkmem );

    typedef int
        (*lzo_decompress)         ( const lzo_bytep src, lzo_uint  src_len,
                                    lzo_bytep dst, lzo_uintp dst_len,
                                    lzo_voidp wrkmem /* NOT USED */ );
    struct LzoType
    {
        const lzo_compress compressor;
        const lzo_decompress decompressor;
        LzoType(lzo_compress compressor, lzo_decompress decompressor) :
            compressor(compressor),
            decompressor(decompressor)
        {}
    };
    static map<const string, LzoType*> LZO_TYPES;
    static const map<const string, LzoType*>& lzo_types()
    {
        if(LZO_TYPES.empty())
        {
            LZO_TYPES = map<const string, LzoType*>();
            LZO_TYPES["1"]       = new LzoType(lzo1_compress,       lzo1_decompress);
            LZO_TYPES["1_99"]    = new LzoType(lzo1_99_compress,    lzo1_decompress);
            LZO_TYPES["1b1"]     = new LzoType(lzo1b_1_compress,    lzo1b_decompress);
            LZO_TYPES["1b2"]     = new LzoType(lzo1b_2_compress,    lzo1b_decompress);
            LZO_TYPES["1b3"]     = new LzoType(lzo1b_3_compress,    lzo1b_decompress);
            LZO_TYPES["1b4"]     = new LzoType(lzo1b_4_compress,    lzo1b_decompress);
            LZO_TYPES["1b5"]     = new LzoType(lzo1b_5_compress,    lzo1b_decompress);
            LZO_TYPES["1b6"]     = new LzoType(lzo1b_6_compress,    lzo1b_decompress);
            LZO_TYPES["1b7"]     = new LzoType(lzo1b_7_compress,    lzo1b_decompress);
            LZO_TYPES["1b8"]     = new LzoType(lzo1b_8_compress,    lzo1b_decompress);
            LZO_TYPES["1b9"]     = new LzoType(lzo1b_9_compress,    lzo1b_decompress);
            LZO_TYPES["1b99"]    = new LzoType(lzo1b_99_compress,   lzo1b_decompress);
            LZO_TYPES["1b999"]   = new LzoType(lzo1b_999_compress,  lzo1b_decompress);
            LZO_TYPES["1c1"]     = new LzoType(lzo1c_1_compress,    lzo1c_decompress);
            LZO_TYPES["1c2"]     = new LzoType(lzo1c_2_compress,    lzo1c_decompress);
            LZO_TYPES["1c3"]     = new LzoType(lzo1c_3_compress,    lzo1c_decompress);
            LZO_TYPES["1c4"]     = new LzoType(lzo1c_4_compress,    lzo1c_decompress);
            LZO_TYPES["1c5"]     = new LzoType(lzo1c_5_compress,    lzo1c_decompress);
            LZO_TYPES["1c6"]     = new LzoType(lzo1c_6_compress,    lzo1c_decompress);
            LZO_TYPES["1c7"]     = new LzoType(lzo1c_7_compress,    lzo1c_decompress);
            LZO_TYPES["1c8"]     = new LzoType(lzo1c_8_compress,    lzo1c_decompress);
            LZO_TYPES["1c9"]     = new LzoType(lzo1c_9_compress,    lzo1c_decompress);
            LZO_TYPES["1c99"]    = new LzoType(lzo1c_99_compress,   lzo1c_decompress);
            LZO_TYPES["1c999"]   = new LzoType(lzo1c_999_compress,  lzo1c_decompress);
            LZO_TYPES["1f1"]     = new LzoType(lzo1f_1_compress,    lzo1f_decompress);
            LZO_TYPES["1f999"]   = new LzoType(lzo1f_999_compress,  lzo1f_decompress);
            LZO_TYPES["1x1"]     = new LzoType(lzo1x_1_compress,    lzo1x_decompress);
            LZO_TYPES["1x1_11"]  = new LzoType(lzo1x_1_11_compress, lzo1x_decompress);
            LZO_TYPES["1x1_12"]  = new LzoType(lzo1x_1_12_compress, lzo1x_decompress);
            LZO_TYPES["1x1_15"]  = new LzoType(lzo1x_1_15_compress, lzo1x_decompress);
            LZO_TYPES["1x1_999"] = new LzoType(lzo1x_999_compress,  lzo1x_decompress);
            LZO_TYPES["1y1"]     = new LzoType(lzo1y_1_compress,    lzo1y_decompress);
            LZO_TYPES["1y999"]   = new LzoType(lzo1y_999_compress,  lzo1y_decompress);
            LZO_TYPES["1z999"]   = new LzoType(lzo1z_999_compress,  lzo1z_decompress);
            LZO_TYPES["2a999"]   = new LzoType(lzo2a_999_compress,  lzo2a_decompress);
        }
        return LZO_TYPES;
    }

    static const LzoType* get_type(const string& args)
    {
        string args_lower = args;
        transform(args_lower.begin(), args_lower.end(), args_lower.begin(), ::tolower);

        return lzo_types().find(args_lower)->second;
    }
    static size_t compressor(char*in, size_t isize, char*out, size_t osize, void* work)
    {
        lzo_uint ret;
        static_lzo_compressor((uint8_t*)in, isize, (uint8_t*)out, &ret, *(char**)work);
        return ret;
    }
    static size_t decompressor(char*in, size_t isize, char*out, size_t osize, void* work)
    {
        lzo_uint ret;
        static_lzo_decompressor((uint8_t*)in, isize, (uint8_t*)out, &ret, *(char**)work);
        return ret;
    }
    static string default_args(const string& args)
    {
        return args == "" ? "1x1" : args;
    }

    char** work_ptrs;
    char * work;

    static lzo_compress static_lzo_compressor;     // set by init() //FIXME: send it as a parameter instead of using statics
    static lzo_decompress static_lzo_decompressor; // set by init()

public:
    LZO() :
        Codec("LZO", _LZO_VERSION, compressor, decompressor, no_blowup),
        work_ptrs(0),
        work(0)
    {}
    virtual void init(const string& args, unsigned threads_no, size_t isize, bool init_compressor, bool init_decompressor)
    {
        int ret = lzo_init();
        if(ret != LZO_E_OK)
            cerr<<"LZO refuses to initialise: " << ret << endl, exit(1);
        if(init_compressor)
        {
            this->work      = new(nothrow) char[LZO1B_999_MEM_COMPRESS*threads_no]();
            this->work_ptrs = new(nothrow) char*[threads_no];
            if (!this->work || !this->work_ptrs)
                cerr<<"Out of memory\n", exit(1);
            for(unsigned i = 0; i < threads_no; ++i)
                this->work_ptrs[i] = this->work + LZO1B_999_MEM_COMPRESS * i;
        }
        this->args = default_args(args);

        const LzoType* t = get_type(this->args);

        static_lzo_compressor   = t->compressor;
        static_lzo_decompressor = t->decompressor;
    }
    virtual void cleanup()
    {
        if(this->work_ptrs)
        {
            delete[] this->work_ptrs;
            delete[] this->work;
            this->work_ptrs = 0;
            this->work = 0;
        }
    }
    string help() const
    {
        const map<const string, LzoType*>& types = lzo_types();
        string ret = this->name + " " + this->version + "\nAvaivable parameters: \n";
        for(map<const string, LzoType*>::const_iterator it = types.begin(); it != types.end(); ++it)
        {
            ret += " ";
            ret += it->first;
            ret += "\n";
        }
        return ret;
    }
    void** eparams()
    {
        return (void**)this->work_ptrs;
    }
    void** dparams()
    {
        return eparams();
    }

};
LZO::lzo_compress LZO::static_lzo_compressor = 0;
LZO::lzo_decompress LZO::static_lzo_decompressor = 0;
map<const string, LZO::LzoType*> LZO::LZO_TYPES;

#endif// FSBENCH_USE_LZO

#ifdef FSBENCH_USE_LZSS_IM

#include "lzss.hpp"

struct LZSS_IM : Codec
{
    struct Params
    {
        LZSSIM coder;
        int mode;
    };
    static size_t compressor(char* in, size_t isize, char* out, size_t osize, void* p)
    {
        Params* params = *(Params**)p;
        return params->coder.lzssim_encode((unsigned char*)in, (unsigned char*)out, isize,  params->mode);
    }
    static size_t decompressor(char* in, size_t isize, char* out, size_t osize, void* _)
    {
        LZSSIM::lzssim_decode((unsigned char*)in, (unsigned char*)out, osize);
        return osize;
    }
    unsigned threads_no;
public:

    LZSS_IM() :
        Codec("LZSS-IM", _LZSSIM_VERSION, compressor, decompressor),
        params(0)
    {}
    virtual void init(const string& args, unsigned threads_no, size_t isize, bool init_compressor, bool init_decompressor)
    {
        this->threads_no = threads_no;
        this->args = args;
        if(init_compressor)
        {
            this->params = (void**)new Params*[threads_no];
            for(unsigned i=0; i<threads_no; ++i)
            {
                Params* p = new Params;
                p->mode = args == "x" || args == "X";
                this->params[i] = p;
            }
        }
    }
    virtual void cleanup()
    {
        if(this->params)
        {
            for(unsigned i=0; i<this->threads_no; ++i)
            {
                delete ((Params**)this->params)[i];
            }
            delete[] this->params;
            this->params = 0;
        }
    }
    void** eparams()
    {
        return params;
    }
    void** dparams()
    {
        return eparams();
    }

protected:
    void** params;
};
#endif //FSBENCH_USE_LZSS_IM

#ifdef FSBENCH_USE_LZV1
extern "C"
{
    #include "lzv1.h"
}

struct LZV1 : Codec
{
    static size_t compressor(char* in, size_t isize, char* out, size_t osize, void* memory)
    {
        return wLZV1((uch*)in, (uch*)out, *(ush**)memory, isize, osize);
    }
    static size_t decompressor(char* in, size_t isize, char* out, size_t osize, void* memory)
    {
        return rLZV1((uch*)in, (uch*)out, isize, osize);
    }
    unsigned threads_no;
public:

    LZV1() :
        Codec("LZV1", _LZV1_VERSION, compressor, decompressor, no_blowup),
        params(0)
    {}
    virtual void init(const string& args, unsigned threads_no, size_t isize, bool init_compressor, bool init_decompressor)
    {
        if(init_compressor)
        {
            this->threads_no = threads_no;
            this->params = new void*[threads_no];
            for(unsigned i=0; i<threads_no; ++i)
            {
                this->params[i] = new ush[16384];
            }
        }
    }
    virtual void cleanup()
    {
        if(this->params)
        {
            for(unsigned i=0; i<this->threads_no; ++i)
            {
                delete[] ((ush**)this->params)[i];
            }
            delete[] this->params;
        }
    }
    void** eparams()
    {
        return params;
    }

protected:
    void** params;
};
#endif //FSBENCH_USE_LZV1

#ifdef FSBENCH_USE_NRV
extern "C"
{
    #include "ucl/ucl.h"
}
struct Nrv : Codec
{
private:
#define NUMBER_OF_NRV_TYPES 3

    typedef int (*nrv_compress) ( const ucl_bytep src, ucl_uint src_len,
                                  ucl_bytep dst, ucl_uintp dst_len,
                                  ucl_progress_callback_p cb,
                                  int level,
                                  const struct ucl_compress_config_p conf,
                                  ucl_uintp result );
    typedef int (*nrv_decompress) ( const ucl_bytep src, ucl_uint  src_len,
                                    ucl_bytep dst, ucl_uintp dst_len,
                                    ucl_voidp wrkmem );
    struct NrvType
    {
        const string name;
        nrv_compress compressor;
        nrv_decompress decompressor;
        NrvType(const string& name, nrv_compress compressor, nrv_decompress decompressor) :
            name(name),
            compressor(compressor),
            decompressor(decompressor)
        {}
    };
    static const NrvType nrv_types[];
    static const NrvType* current_type;//set by init()  //FIXME: send it as a parameter instead of using statics

    static const string default_mode;
    const string& default_args(const string& args)
    {
        return args == "" ? default_mode : args;
    }

    static size_t compressor(char* in, size_t isize, char* out, size_t osize, void* mode)
    {
        ucl_uint ret;
        current_type->compressor((const unsigned char*)in, isize, (unsigned char*)out, &ret, NULL, *(uintptr_t*)mode, NULL, NULL);
        return (int)ret;
    }
    static size_t decompressor(char* in, size_t isize, char* out, size_t osize, void* _)
    {
        ucl_uint ret;
        current_type->decompressor((const unsigned char*)in, isize, (unsigned char*)out, &ret, NULL);
        return (int)ret;
    }
    static const string full_name(const string& type)
    {
        return string("nrv2")+type;
    }
    const NrvType* get_type(const string& type_name)
    {
        for(int i=0; i<NUMBER_OF_NRV_TYPES; ++i)
        {
            if(case_insensitive_compare(nrv_types[i].name, type_name) == 0)
                return &nrv_types[i];
        }
        throw InvalidParams(help());
    }

public:

    Nrv(const string& type) :
        Codec(full_name(type), _NRV_VERSION, compressor, decompressor, no_blowup),
        params(0)
    {}
    virtual void init(const string& args, unsigned threads_no, size_t isize, bool init_compressor, bool init_decompressor)
    {
        if(init_compressor)
        {
            this->args = default_args(args); // I let the decoder ignore args, they are not used and they won't be printed in the output
            uintptr_t mode;
            try
            {
                from_string(this->args, mode);
            }
            catch(const bad_cast &bc)
            {
                throw InvalidParams(help());
            }
            
            if(mode < 1 || mode > 10)
                throw InvalidParams(help());
            this->params = new void*[threads_no];
            for(unsigned i=0; i<threads_no; ++i)
            {
                this->params[i] = (void*)mode;
            }
        }
        this->current_type = get_type(this->name);
    }
    virtual void cleanup()
    {
        delete[] this->params;
    }

    virtual string help() const
    {
        return this->name + " " + this->version + "\nAvaivable parameters: \n Numbers from 1 to 10\n";
    }
    
    void** eparams()
    {
        return params;
    }

protected:
    void** params;
};
const Nrv::NrvType Nrv::nrv_types[] = {
    Nrv::NrvType("nrv2b", ucl_nrv2b_99_compress, ucl_nrv2b_decompress_8),
    Nrv::NrvType("nrv2d", ucl_nrv2d_99_compress, ucl_nrv2d_decompress_8),
    Nrv::NrvType("nrv2e", ucl_nrv2e_99_compress, ucl_nrv2e_decompress_8)
};
const Nrv::NrvType* Nrv::current_type = 0;
const string Nrv::default_mode("6");

#endif// FSBENCH_USE_NRV

#ifdef FSBENCH_USE_QUICKLZ

extern "C"
{
    #include "quicklz/quicklz.h"
}

struct QuickLZ : Codec
{
private:

    typedef size_t (*qlz_compressor)  (const void *source, char *destination, size_t size, qlz_state_compress *state);
    typedef size_t (*qlz_decompressor)(const char *source, void *destination, qlz_state_decompress *state);

    struct QuickLZType
    {
        const qlz_compressor compressor;
        const qlz_decompressor decompressor;
        QuickLZType(qlz_compressor compressor, qlz_decompressor decompressor) :
            compressor(compressor),
            decompressor(decompressor)
        {}
    };
    static map<const string, QuickLZType*> QLZ_TYPES;
    static const map<const string, QuickLZType*>& qlz_types()
    {
        if(QLZ_TYPES.empty())
        {
            QLZ_TYPES = map<const string, QuickLZType*>();
            QLZ_TYPES["1"] = new QuickLZType(qlz_compress,                    qlz_decompress);
            QLZ_TYPES["2"] = new QuickLZType(qlz_compress2,                   qlz_decompress2);
            QLZ_TYPES["3"] = new QuickLZType(qlz_compress3, (qlz_decompressor)qlz_decompress3);
        }
        return QLZ_TYPES;
    }

    static const QuickLZType* get_type(const string& args)
    {
        string args_lower = args;
        transform(args_lower.begin(), args_lower.end(), args_lower.begin(), ::tolower);

        return qlz_types().find(args_lower)->second;
    }
    static size_t compressor(char* in, size_t isize, char* out, size_t osize, void* work)
    {
        return static_qlz_compressor(in,out,isize,*(qlz_state_compress**)work);
    }
    static size_t decompressor(char* in, size_t isize, char* out, size_t osize, void* work)
    {
        return static_qlz_decompressor(in,out,*(qlz_state_decompress**)work);
    }
    static string default_args(const string& args)
    {
        return args == "" ? "1" : args;
    }

    static qlz_compressor static_qlz_compressor; // set by init()  //FIXME: send it as a parameter instead of using statics
    static qlz_decompressor static_qlz_decompressor; // set by init()

public:
    qlz_state_compress**   qlz_compression_states;
    qlz_state_decompress** qlz_decompression_states;
    unsigned threads_no;

    QuickLZ() :
        Codec("QuickLZ", _QLZ_VERSION, compressor, decompressor),
        qlz_compression_states(0),
        qlz_decompression_states(0)
    {}

    virtual void init(const string& args, unsigned threads_no, size_t isize, bool init_compressor, bool init_decompressor)
    {        
        if(init_compressor)
        {
            this->qlz_compression_states = new(nothrow) qlz_state_compress*[threads_no];
            if (!this->qlz_compression_states)
                cerr<<"Out of memory\n", exit(1);
                
            int state_size = max(qlz_get_setting3(1), max(qlz_get_setting2(1), qlz_get_setting(1)));
            
            for(unsigned i = 0; i < threads_no; ++i)
            {
                this->qlz_compression_states[i] = (qlz_state_compress*) new(nothrow) char[state_size] ();
                if (!this->qlz_compression_states[i])
                    cerr<<"Out of memory\n", exit(1);
            }
        }
        if(init_decompressor)
        {
            this->qlz_decompression_states = new(nothrow) qlz_state_decompress*[threads_no];
            if (!this->qlz_decompression_states)
                cerr<<"Out of memory\n", exit(1);
                
            int dstate_size = max(qlz_get_setting3(2), max(qlz_get_setting2(2), qlz_get_setting(2)));
            
            for(unsigned i = 0; i < threads_no; ++i)
            {
                this->qlz_decompression_states[i] = (qlz_state_decompress*) new(nothrow) char[dstate_size] ();
                if (!this->qlz_decompression_states[i])
                    cerr<<"Out of memory\n", exit(1);
            }
        }
        this->threads_no = threads_no;

        this->args = default_args(args);

        const QuickLZType* t = get_type(this->args);

        static_qlz_compressor   = t->compressor;
        static_qlz_decompressor = t->decompressor;
    }
    virtual void cleanup()
    {
        if(this->qlz_compression_states)
        {
            for(unsigned i = 0; i < this->threads_no; ++i)
            {
                delete[] this->qlz_compression_states[i];
            }
            delete[] this->qlz_compression_states;
            this->qlz_compression_states = 0;
        }
        if(this->qlz_decompression_states)
        {
            for(unsigned i = 0; i < this->threads_no; ++i)
            {
                delete[] this->qlz_decompression_states[i];
            }
            delete[] this->qlz_decompression_states;
            this->qlz_decompression_states = 0;
        }
    }
    string help() const
    {
        const map<const string, QuickLZType*>& types = qlz_types();
        string ret = this->name + " " + this->version + "\nAvaivable parameters: \n";
        for(map<const string, QuickLZType*>::const_iterator it = types.begin(); it != types.end(); ++it)
        {
            ret += " ";
            ret += it->first;
            ret += "\n";
        }
        return ret;
    }
    void** eparams()
    {
        return (void**)this->qlz_compression_states;
    }
    void** dparams()
    {
        return (void**)this->qlz_decompression_states;
    }
};
QuickLZ::qlz_compressor QuickLZ::static_qlz_compressor = 0;
QuickLZ::qlz_decompressor QuickLZ::static_qlz_decompressor = 0;
map<const string, QuickLZ::QuickLZType*> QuickLZ::QLZ_TYPES;

#endif //FSBENCH_USE_QUICKLZ

#ifdef FSBENCH_USE_TOR

#include "tornado/tor_test.h"

struct Tor : Codec
{
private:

    static const string default_mode;
    const string& default_args(const string& args)
    {
        return args == "" ? default_mode : args;
    }
    static size_t compressor(char* in, size_t isize, char* out, size_t osize, void* mode)
    {
        return tor_compress(*(uintptr_t*)mode, (uint8_t*)in, (uint8_t*)out, isize);
    }
    static size_t decompressor(char* in, size_t isize, char* out, size_t osize, void* _)
    {
        return tor_decompress((uint8_t*)in, (uint8_t*)out, isize);
    }

public:

    Tor() :
        Codec("tornado", _TOR_VERSION, compressor, decompressor),
        params(0)
    {}
    virtual void init(const string& args, unsigned threads_no, size_t isize, bool init_compressor, bool init_decompressor)
    {
        if(init_compressor)
        {
            this->args = default_args(args);
            uintptr_t mode;
            try
            {
                from_string(this->args, mode);
            }
            catch(const bad_cast &bc)
            {
                throw InvalidParams(help());
            }
            
            if(mode < 1 || mode > 11)
                throw InvalidParams(help());
            this->params = new void*[threads_no];
            for(unsigned i=0; i<threads_no; ++i)
            {
                this->params[i] = (void*)mode;
            }
        }
    }
    virtual void cleanup()
    {
        if(this->params)
            delete[] this->params;
    }

    virtual string help() const
    {
        return this->name + " " + this->version + "\nAvaivable parameters: \n Numbers from 1 to 11\n";
    }
    
    void** eparams()
    {
        return params;
    }

protected:
    void** params;
};
const string Tor::default_mode("6");

#endif// FSBENCH_USE_TOR

#ifdef FSBENCH_USE_YAPPY

#include "Yappy.hpp"

struct Yappy : CodecWithIntModes
{
private:
    static size_t compressor(char* in, size_t isize, char* out, size_t osize, void* mode)
    {
        return YappyCompress((ui8*)in, (ui8*)out, isize, *(uintptr_t*)mode) - (ui8*)out;
    }
    static size_t decompressor(char* in, size_t isize, char* out, size_t osize, void* _)
    {
        return YappyUnCompress((ui8*)in, (ui8*)(in+isize), (ui8*)out) - (ui8*)out;
    }
public:
    Yappy():
        CodecWithIntModes("Yappy", _YAPPY_VERSION, compressor, decompressor, 1, 256, "10")
    {
        YappyFillTables();
    }
};
#endif// FSBENCH_USE_YAPPY

/////////////////////////////////
// Global variables 
// and helpers to manipulate them
/////////////////////////////////

// Just looks up a codec in CODECS
Codec* raw_find_codec(const string& name)
{
    for(list<Codec*>::iterator it=CODECS.begin(); it != CODECS.end(); ++it)
    {
        if(!case_insensitive_compare((*it)->name, name))
        {
            return *it;
        }
    }
    return 0;
};

// Looks up a codec in CODECS.
// When it can't be found, but is a combination of other codecs,
// it adds it to CODECS
// The addition thing is here because it is needed to create a new codec
// But who is supposed to dispose it?
// Not the caller because it can't tell whether the pointer shows something created now or before
// The could be worked around, but I think that only in some ugly ways
// Not we, because the codec wouldn't live long enough
// This has some funny side effects like codec1/codec2/codec3 that works iff you did codec2/codec3 before
Codec* find_codec(const string& name)
{
    Codec* ret = raw_find_codec(name);
    if(!ret)
    {
        // maybe it's a combination of existing codecs?
        string::size_type pos = name.find('/');
        if(pos != string::npos)
        {
            string encoder_name = name.substr(0, pos);
            string decoder_name = name.substr(pos + 1);
            Codec* encoder = raw_find_codec(encoder_name);
            Codec* decoder = raw_find_codec(decoder_name);
            if(encoder && decoder)
            {
                // good, it's a combination of existing codecs
                CombinationCodec* codec = new CombinationCodec(*encoder, *decoder);
                CODECS.push_back(codec);
                ret = codec;
            }
        }
    }
    return ret;
};

vector<string> pretty_list_of_codecs()
{
    vector<string> ret;
    for(list<Codec*>::iterator it=CODECS.begin(); it != CODECS.end(); ++it)
    {
        const string& name = (*it)->name;
        if(name.find('/') == string::npos)
            ret.push_back(name);
    }
    sort(ret.begin(), ret.end());
    return ret;
}

#ifdef FSBENCH_USE_FASTLZ
static const CodecArgs fastlz_args[] = {
    CodecArgs("1", fastlz1_c, fastlz_d),
    CodecArgs("2", fastlz2_c, fastlz_d)};
#endif
#ifdef FSBENCH_USE_LZ4
static const CodecArgs LZ4_args[] = {
    CodecArgs("12",      LZ4_c,         LZ4_d),
    CodecArgs("17",      LZ4_17_c,      LZ4_d)};
#endif
#ifdef FSBENCH_USE_LZF
static const CodecArgs LZF_args[] = {
    CodecArgs("",      LZF_c,       LZF_d),
    CodecArgs("very",  LZF_very_c,  LZF_d),
    CodecArgs("ultra", LZF_ultra_c, LZF_d)
};
#endif
#ifdef FSBENCH_USE_RLE64
static const CodecArgs RLE64_args[] = {
    CodecArgs("8",  RLE8_c,  RLE8_d),
    CodecArgs("16", RLE16_c, RLE16_d),
    CodecArgs("32", RLE32_c, RLE32_d),
    CodecArgs("64", RLE64_c, RLE64_d)
};
#endif

Codec* codecs[] =
{
#ifdef FSBENCH_USE_FASTLZ
    new MultifunctionCodec(
        "fastlz", _FASTLZ_VERSION,
        fastlz_args,
        2,
        "1",
        fastlz_m),
#endif
#ifdef FSBENCH_USE_LZ4
    new MultifunctionCodec(
        "LZ4", _LZ4_VERSION,
        LZ4_args,
        2,
        "12",
        LZ4_m),
    new Codec("LZ4hc",    _LZ4HC_VERSION,    LZ4hc_c,    LZ4_d,      LZ4_m),
#endif
#ifdef FSBENCH_USE_LZF
    new MultifunctionCodec(
        "LZF", _LZF_VERSION,
        LZF_args,
        3,
        "",
        no_blowup),
#endif
#ifdef FSBENCH_USE_RLE64
    new MultifunctionCodec(
        "RLE64", _RLE64_VERSION,
        RLE64_args,
        4,
        "64"),
#endif
#ifdef FSBENCH_USE_7Z
    new CodecWithIntModes("7z-deflate",       _7z_VERSION,       FsBench7z::deflate,       FsBench7z::inflate,       1,        9,        "5", no_blowup),
    new CodecWithIntModes("7z-deflate64",     _7z_VERSION,       FsBench7z::deflate64,     FsBench7z::inflate64,     1,        9,        "5", no_blowup),
#endif
#ifdef FSBENCH_USE_BLOSC
    new CodecWithIntModes("blosc",            _BLOSC_VERSION,    blosc_c,                  blosc_d,                   1,        9,        "5", no_blowup),
#endif
#ifdef FSBENCH_USE_BZ2
    new CodecWithIntModes("bzip2",            _BZ2_VERSION,      bz2_c,                    bz2_d,                     1,        9,        "9", no_blowup),
#endif
#ifdef FSBENCH_USE_CRYPTOPP
    new CodecWithIntModes("cryptopp-deflate", _CRYPTOPP_VERSION, FsBenchCryptoPP::deflate, FsBenchCryptoPP::inflate, 1,        9,        "6", no_blowup),
#endif
#ifdef FSBENCH_USE_LODEPNG
    new CodecWithIntModes("lodepng",          _LODEPNG_VERSION,  lodepng_c,                lodepng_d,                64,    32768,    "32768", no_blowup),
#endif
#ifdef FSBENCH_USE_LZG 
    new CodecWithIntModes("lzg",              _LZG_VERSION,      LZG_c,                    LZG_d,                     1,        9,        "5", LZG_m),
#endif
#ifdef FSBENCH_USE_LZX_COMPRESS
    new CodecWithIntModes("lzx_compress",     _LZX_VERSION,      LZX_c,                    0,                        15,       21,       "21"),
#endif
#ifdef FSBENCH_USE_MINIZ
    new CodecWithIntModes("miniz",            _MINIZ_VERSION,    miniz_c,                  miniz_d,                   1,       10,        "6", no_blowup),
#endif
#ifdef FSBENCH_USE_ZLIB
    new CodecWithIntModes("zlib",             _ZLIB_VERSION,     zlib_c,                   zlib_d,                    1,        9,        "6", no_blowup),
#endif
#if 0
#ifdef FSBENCH_USE_7Z
    new Codec("7z-lzx",          _7z_VERSION,       0,                    FsBench7z::unlzx,      no_blowup),
#endif
#endif
#ifdef FSBENCH_USE_BCL
    new Codec("bcl-huffman",     _BCL_VERSION,      FsBenchBCL::huffman,  FsBenchBCL::unhuffman, FsBenchBCL::huffman_maxsize),
    new Codec("bcl-lz",          _BCL_VERSION,      FsBenchBCL::lz,       FsBenchBCL::unlz,      FsBenchBCL::lz_maxsize),
    new Codec("bcl-rle",         _BCL_VERSION,      FsBenchBCL::rle,      FsBenchBCL::unrle,     FsBenchBCL::rle_maxsize),
    new FsBenchBCL::LZFast(),
#endif
#ifdef FSBENCH_USE_DOBOZ
    new Codec("Doboz",           _DOBOZ_VERSION,    Doboz_c,              Doboz_d,                no_blowup),
#endif
#ifdef FSBENCH_USE_HALIBUT
    new Codec("Halibut-deflate", _HALIBUT_VERSION,  halibut_c,            halibut_d,              no_blowup),
#endif
#ifdef FSBENCH_USE_LZFX
    new Codec("LZFX",            _LZFX_VERSION,     LZFX_c,               LZFX_d),
#endif
#ifdef FSBENCH_USE_LZJB
    new Codec("lzjb",            _LZJB_VERSION,     LZJB_c,               LZJB_d,                 no_blowup),
#endif
#ifdef FSBENCH_USE_LZMAT
    new Codec("lzmat",           _LZMAT_VERSION,    lzmat_c,              lzmat_d,                no_blowup),
#endif
#ifdef FSBENCH_USE_QUICKLZZIP
    new Codec("QuickLZ-zip",     _QLZZIP_VERSION,   qlzzip_c,             0),
#endif
#ifdef FSBENCH_USE_SHRINKER
    new Codec("Shrinker",        _SHRINKER_VERSION, Shrinker_c,           Shrinker_d,             Shrinker_m),
#endif
#ifdef FSBENCH_USE_SNAPPY
    new Codec("Snappy",          _SNAPPY_VERSION,   snappy_c,             snappy_d,               snappy_m),
#endif
#ifdef FSBENCH_USE_TINF
    new Codec("tinf",            _TINF_VERSION,     0,                    FsBenchTinf::inflate,   no_blowup),
#endif
#ifdef FSBENCH_USE_NRV
    new Nrv("b"),
    new Nrv("d"),
    new Nrv("e"),
#endif
#ifdef FSBENCH_USE_BLZ
    new BriefLZ(),
#endif
#ifdef FSBENCH_USE_LZHAM
    new LZHAM(),
#endif
#ifdef FSBENCH_USE_LZO
    new LZO(),
#endif
#ifdef FSBENCH_USE_LZSS_IM
    new LZSS_IM(),
#endif
#ifdef FSBENCH_USE_LZV1
    new LZV1(),
#endif
#ifdef FSBENCH_USE_QUICKLZ
    new QuickLZ(),
#endif
#ifdef FSBENCH_USE_SCZ
    new Scz(),
#endif
#ifdef FSBENCH_USE_TOR
    new Tor(),
#endif
#ifdef FSBENCH_USE_YAPPY
    new Yappy,
#endif
#ifdef FSBENCH_USE_CITYHASH
    new Checksum<  sizeof(uint64_t)>("CityHash64",  _CITYHASH_VERSION, CityHash64),
    new Checksum<2*sizeof(uint64_t)>("CityHash128", _CITYHASH_VERSION, CityHash128),
#endif
#ifdef FSBENCH_USE_CRAPWOW
    new Checksum<sizeof(unsigned int)>("CrapWow", _CRAPWOW_VERSION, CrapWow),
#endif
#ifdef FSBENCH_USE_CRYPTOPP
    new Checksum<sizeof(uint32_t)>("cryptopp-adler32", _CRYPTOPP_VERSION, FsBenchCryptoPP::adler32),
    new Checksum<sizeof(uint32_t)>("cryptopp-crc32",   _CRYPTOPP_VERSION, FsBenchCryptoPP::crc),
    new Checksum<160/CHAR_BIT>    ("cryptopp-md5",     _CRYPTOPP_VERSION, FsBenchCryptoPP::md5),
    new Checksum<224/CHAR_BIT>    ("cryptopp-sha224",  _CRYPTOPP_VERSION, FsBenchCryptoPP::sha224),
    new Checksum<256/CHAR_BIT>    ("cryptopp-sha256",  _CRYPTOPP_VERSION, FsBenchCryptoPP::sha256),
    new Checksum<384/CHAR_BIT>    ("cryptopp-sha384",  _CRYPTOPP_VERSION, FsBenchCryptoPP::sha384),
    new Checksum<512/CHAR_BIT>    ("cryptopp-sha512",  _CRYPTOPP_VERSION, FsBenchCryptoPP::sha512),
#endif
#ifdef FSBENCH_USE_MURMUR
    new Checksum<  sizeof(uint32_t)>("murmur3_x86_32",  _MURMUR_VERSION, murmur_x86_32),
    new Checksum<2*sizeof(uint64_t)>("murmur3_x86_128", _MURMUR_VERSION, murmur_x86_128),
    new Checksum<2*sizeof(uint64_t)>("murmur3_x64_128", _MURMUR_VERSION, murmur_x64_128),
#endif
#ifdef FSBENCH_USE_SPOOKY
    new Checksum<2*sizeof(uint64_t)>("SpookyHash", _SPOOKY_VERSION, spooky),
#endif
#ifdef FSBENCH_USE_SANMAYCE_FNV
    new Checksum<sizeof(uint64_t)>("fnv1a-tesla",     _SANMAYCE_VERSION, fnv1_tesla),
    new Checksum<sizeof(uint64_t)>("fnv1a-tesla3",     _SANMAYCE_VERSION, fnv1_tesla3),
    new Checksum<sizeof(uint32_t)>("fnv1a-yoshimura", _SANMAYCE_VERSION, fnv1_yoshimura),
    new Checksum<sizeof(uint32_t)>("fnv1a-jesteress", _SANMAYCE_VERSION, fnv1_jesteress),
    new Checksum<sizeof(uint32_t)>("fnv1a-meiyan",    _SANMAYCE_VERSION, fnv1_meiyan),
    new Checksum<sizeof(uint32_t)>("fnv1a-mantis",    _SANMAYCE_VERSION, fnv1_mantis),
#endif
#ifdef FSBENCH_USE_XXHASH
    new Checksum<sizeof(unsigned int)>("xxhash-fast",      _XXHASH_VERSION, xxhash_fast),
    new Checksum<sizeof(unsigned int)>("xxhash-strong",    _XXHASH_VERSION, xxhash_strong),
    new Checksum<4*sizeof(uint64_t)>  ("xxhash256",   _XXHASH_VERSION, xxhash_256),
#endif
#ifdef FSBENCH_USE_ZFS
    new Checksum<4*sizeof(uint64_t)>("fletcher2", _ZFS_VERSION, fletcher2),
    new Checksum<4*sizeof(uint64_t)>("fletcher4", _ZFS_VERSION, fletcher4),
#endif
#ifdef FSBENCH_USE_ECRYPT
    new FsBenchECrypt::AES128Bernstein(),
    new FsBenchECrypt::AES256Hongjun(),
    new FsBenchECrypt::ChaCha(),
    new FsBenchECrypt::HC128(),
    new FsBenchECrypt::HC256(),
    new FsBenchECrypt::Lex(),
    new FsBenchECrypt::Rabbit(),
    new FsBenchECrypt::RC4(),
    new FsBenchECrypt::Salsa20_8(),
    new FsBenchECrypt::Salsa20_12(),
    new FsBenchECrypt::Salsa20(),
    new FsBenchECrypt::Snow2(),
    new FsBenchECrypt::Sosemanuk(),
    new FsBenchECrypt::Trivium(),
#endif
    new Codec("nop",      "0",               nop_c,      nop_d,      no_blowup)
};

list<Codec*> CODECS = list<Codec*>(codecs, codecs + sizeof(codecs) / sizeof(Codec*));

list<CodecWithParams> createParamsList(list<pair<Codec*, const string> > lst)
{
    list<CodecWithParams> ret;
    for(list<pair<Codec*, const string> >::iterator it=lst.begin(); it != lst.end(); ++it)
    {
        if(it->first)
            ret.push_back(CodecWithParams(*(it->first), it->second));
    }
    return ret;
}
#define MKLIST(name, array) list<CodecWithParams> (name) = createParamsList(list< pair<Codec*,const string> >((array), (array) + sizeof((array)) / sizeof(pair<Codec*,const string>)))

pair<Codec*,string> default_codecs[] =
{
    make_pair(raw_find_codec("LZ4"),     ""),
    make_pair(raw_find_codec("LZO"),     ""),
    make_pair(raw_find_codec("QuickLZ"), ""),
    make_pair(raw_find_codec("Snappy"),  "")
};

MKLIST(DEFAULT_CODECS, default_codecs);

pair<Codec*,string> fast_codecs[] =
{
    make_pair(raw_find_codec("bcl-rle"),  ""),
    make_pair(raw_find_codec("blosc"),    ""),
    make_pair(raw_find_codec("fastlz"),   ""),
    make_pair(raw_find_codec("LZ4"),      ""),
    make_pair(raw_find_codec("LZF"),      "very"),
    make_pair(raw_find_codec("LZJB"),     ""),
    make_pair(raw_find_codec("LZO"),      ""),
    make_pair(raw_find_codec("QuickLZ"),  ""),
    make_pair(raw_find_codec("RLE64"),    ""),
    make_pair(raw_find_codec("Shrinker"), ""),
    make_pair(raw_find_codec("Snappy"),   "")
};

MKLIST(FAST_CODECS, fast_codecs);

pair<Codec*,string> all_codecs[] =
{
    make_pair(raw_find_codec("7z-deflate"),      ""),
    make_pair(raw_find_codec("7z-deflate64"),    ""),
    make_pair(raw_find_codec("bcl-huffman"),     ""),
    make_pair(raw_find_codec("bcl-lz"),          ""),
    make_pair(raw_find_codec("bcl-lzfast"),      ""),
    make_pair(raw_find_codec("bcl-rle"),         ""),
    make_pair(raw_find_codec("blosc"),           ""),
    make_pair(raw_find_codec("BriefLZ"),         ""),
    make_pair(raw_find_codec("bzip2"),           ""),
    make_pair(raw_find_codec("cryptopp-deflate"),""),
    make_pair(raw_find_codec("Doboz"),           ""),
    make_pair(raw_find_codec("fastlz"),          ""),
    make_pair(raw_find_codec("halibut-deflate"), ""),
    make_pair(raw_find_codec("lodepng"),         ""),
    make_pair(raw_find_codec("LZ4"),             ""),
    make_pair(raw_find_codec("LZ4hc"),           ""),
    make_pair(raw_find_codec("LZF"),             ""),
    make_pair(raw_find_codec("LZFX"),            ""),
    make_pair(raw_find_codec("lzg"),             ""),
    make_pair(raw_find_codec("LZHAM"),           ""),
    make_pair(raw_find_codec("LZJB"),            ""),
    make_pair(raw_find_codec("lzmat"),           ""),
    make_pair(raw_find_codec("LZO"),             ""),
    make_pair(raw_find_codec("LZSS-IM"),         ""),
    make_pair(raw_find_codec("lzv1"),            ""),
    make_pair(    find_codec("lzx_compress/nop"),""),
    make_pair(raw_find_codec("miniz"),           ""),
    make_pair(raw_find_codec("nrv2b"),           ""),
    make_pair(raw_find_codec("nrv2d"),           ""),
    make_pair(raw_find_codec("nrv2e"),           ""),
    make_pair(raw_find_codec("QuickLZ"),         ""),
    make_pair(raw_find_codec("QuickLZ-zip/zlib"),""),
    make_pair(raw_find_codec("RLE64"),           ""),
    make_pair(raw_find_codec("scz"),             ""),
    make_pair(raw_find_codec("Shrinker"),        ""),
    make_pair(raw_find_codec("Snappy"),          ""),
    make_pair(    find_codec("zlib/tinf"),       ""),
    make_pair(raw_find_codec("Tornado"),         ""),
    make_pair(raw_find_codec("Yappy"),           ""),
    make_pair(raw_find_codec("zlib"),            "")
    //note: nop is not included
};
MKLIST(ALL_CODECS, all_codecs);